# Copyright 2014 Cisco Systems, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os

import contextlib

from oslo_config import cfg
from oslo_log import log as logging
import six
import webob.exc

from neutron.tests.unit.db import test_db_base_plugin_v2
from neutron_lib import exceptions as n_exc

import networking_cisco
from networking_cisco._i18n import _LE
from networking_cisco import backwards_compatibility as bc
from networking_cisco.plugins.cisco.common import (cisco_constants as
                                                   c_constants)
from networking_cisco.plugins.cisco.db.device_manager import (
    hosting_device_manager_db as hdm_db)
from networking_cisco.plugins.cisco.extensions import (
    ciscohostingdevicemanager as ciscodevmgr)
from networking_cisco.plugins.cisco.extensions import ha
from networking_cisco.plugins.cisco.extensions import routertype
from networking_cisco.tests.unit.cisco.device_manager import (
    device_manager_test_support)
from networking_cisco.tests.unit.cisco.device_manager import (
    test_db_device_manager)
from networking_cisco.tests.unit.cisco.l3 import l3_router_test_support

LOG = logging.getLogger(__name__)


policy_path = (os.path.abspath(networking_cisco.__path__[0]) +
               '/../etc/policy.json')

CORE_PLUGIN_KLASS = device_manager_test_support.CORE_PLUGIN_KLASS
L3_PLUGIN_KLASS = l3_router_test_support.L3_PLUGIN_KLASS

NS_ROUTERTYPE_NAME = c_constants.NAMESPACE_ROUTER_TYPE
VM_ROUTERTYPE_NAME = test_db_device_manager.VM_ROUTERTYPE_NAME
HW_ROUTERTYPE_NAME = c_constants.ASR1K_ROUTER_TYPE

NOOP_SCHEDULER = ('networking_cisco.plugins.cisco.l3.schedulers.'
                  'noop_l3_router_hosting_device_scheduler.'
                  'NoopL3RouterHostingDeviceScheduler')
NOOP_RT_DRIVER = ('networking_cisco.plugins.cisco.l3.drivers.'
                  'noop_routertype_driver.NoopL3RouterDriver')
NOOP_AGT_SVC_HELPER = NOOP_SCHEDULER
NOOP_AGT_DRV = NOOP_SCHEDULER

TEST_SLOT_NEED = 2

RT_SETTINGS = {
    NS_ROUTERTYPE_NAME: {
        'slot_need': 0,
        'scheduler': NOOP_SCHEDULER,
        'router_type_driver': NOOP_RT_DRIVER,
        'cfg_agent_driver': NOOP_AGT_DRV},
    VM_ROUTERTYPE_NAME: {
        'slot_need': TEST_SLOT_NEED,
        'scheduler': 'networking_cisco.plugins.cisco.l3.schedulers.'
                     'l3_router_hosting_device_scheduler.'
                     'L3RouterHostingDeviceLongestRunningScheduler',
        'driver': NOOP_RT_DRIVER,
        'cfg_agent_service_helper': NOOP_AGT_SVC_HELPER,
        'cfg_agent_driver': NOOP_AGT_DRV},
    HW_ROUTERTYPE_NAME: {
        'slot_need': 200,
        'scheduler': 'networking_cisco.plugins.cisco.l3.schedulers.'
                     'l3_router_hosting_device_scheduler.'
                     'L3RouterHostingDeviceRandomScheduler',
        'driver': NOOP_RT_DRIVER,
        'cfg_agent_service_helper': NOOP_AGT_SVC_HELPER,
        'cfg_agent_driver': NOOP_AGT_DRV}}


class RoutertypeTestCaseMixin(object):

    def _create_routertype(self, fmt, template_id, name, slot_need,
                           expected_res_status=None, **kwargs):
        data = self._get_test_routertype_attr(template_id=template_id,
                                              name=name, slot_need=slot_need,
                                              **kwargs)
        data.update({'tenant_id': kwargs.get('tenant_id', self._tenant_id)})
        data = {'routertype': data}
        rt_req = self.new_create_request('routertypes', data, fmt)
        if kwargs.get('set_context') and 'tenant_id' in kwargs:
            # create a specific auth context for this request
            rt_req.environ['neutron.context'] = bc.context.Context(
                '', kwargs['tenant_id'])
        hd_res = rt_req.get_response(self.ext_api)
        if expected_res_status:
            self.assertEqual(hd_res.status_int, expected_res_status)
        return hd_res

    @contextlib.contextmanager
    def routertype(self, template_id, name='router type 1',
                   slot_need=TEST_SLOT_NEED, fmt=None, no_delete=False,
                   **kwargs):
        if not fmt:
            fmt = self.fmt
        res = self._create_routertype(fmt, template_id, name, slot_need,
                                      **kwargs)
        if res.status_int >= 400:
            raise webob.exc.HTTPClientError(code=res.status_int)
        routertype = self.deserialize(fmt or self.fmt, res)
        yield routertype
        if not no_delete:
            self._delete('routertypes', routertype['routertype']['id'])

    def _get_test_routertype_attr(self, template_id, name='router type 1',
                                  slot_need=TEST_SLOT_NEED, **kwargs):
        data = {
            'name': name,
            'description': kwargs.get('description'),
            'template_id': template_id,
            'ha_enabled_by_default': kwargs.get('ha_enabled_by_default',
                                                False),
            'shared': kwargs.get('shared', True),
            'slot_need': slot_need,
            'scheduler': kwargs.get('scheduler', NOOP_SCHEDULER),
            'driver': NOOP_RT_DRIVER,
            'cfg_agent_service_helper': NOOP_AGT_SVC_HELPER,
            'cfg_agent_driver': kwargs.get('cfg_agent_driver', NOOP_AGT_DRV)}
        return data

    def _get_non_admin_routertype_attr(self, template_id, name='router type 1',
                                       slot_need=TEST_SLOT_NEED, **kwargs):
        subset = {'id', 'name', 'description', 'tenant_id',
                  'ha_enabled_by_default'}
        attr_dict = self._get_test_routertype_attr(template_id, name,
                                                   slot_need, **kwargs)
        return {kv[0]: kv[1] for kv in attr_dict.items() if kv[0] in subset}

    def _test_list_resources(self, resource, items,
                             neutron_context=None,
                             query_params=None):
        if resource.endswith('y'):
            resource_plural = resource.replace('y', 'ies')
        else:
            resource_plural = resource + 's'

        res = self._list(resource_plural,
                         neutron_context=neutron_context,
                         query_params=query_params)
        resource = resource.replace('-', '_')
        self.assertEqual(sorted([i['id'] for i in res[resource_plural]]),
                         sorted([i[resource]['id'] for i in items]))

    def _test_create_routertypes(self, mappings=None):
        if mappings is None:
            mappings = {}
        for mapping in mappings:
            template = mapping['template']
            if template is None:
                self._routertypes[mapping['router_type']] = None
            else:
                routertype_name = mapping['router_type']
                self._create_routertype(
                    self.fmt, template['hosting_device_template']['id'],
                    routertype_name,
                    RT_SETTINGS[routertype_name]['slot_need'],
                    scheduler=RT_SETTINGS[routertype_name]['scheduler'],
                    ha_enabled_by_default=self._is_ha_tests)

    def _test_remove_routertypes(self, delete_routers=True):
        if delete_routers:
            auto_deleted_router_ids = set()
            routers = self._list('routers')['routers']
            for r in routers:
                # Exclude any redundancy routers as they are removed
                # automatically when removing the user visible router
                for rr_info in r.get(
                        ha.DETAILS,
                        {ha.REDUNDANCY_ROUTERS: []})[ha.REDUNDANCY_ROUTERS]:
                    auto_deleted_router_ids.add(rr_info['id'])
            for r in routers:
                if r['id'] in auto_deleted_router_ids:
                    continue
                # Remove any static routes
                if r.get('routes'):
                    self._update('routers', r['id'],
                                 {'router': {'routes': None}})
                # Remove any floatingips using the router
                for fip in self._list(
                        'floatingips',
                        query_params='router_id=%s' % r['id'])['floatingips']:
                    try:
                        self._delete('floatingips', fip['id'])
                    except n_exc.NeutronException as e:
                        # since this is a cleanup after a UT execution we
                        # just log it and then ignore it. The subsequent router
                        # delete should anyway capture more serious errors.
                        LOG.error(_LE('Encountered error during router clean '
                                      'up: '), e)
                # Remove any router interfaces
                for p in self._list(
                        'ports',
                        query_params='device_id=%s&device_owner=%s' % (
                            r['id'],
                            bc.constants.DEVICE_OWNER_ROUTER_INTF))['ports']:
                    # get_ports can be mocked in some tests so we need to
                    # ensure we get a port that is indeed a router port.
                    try:
                        if (p.get('device_owner') ==
                                bc.constants.DEVICE_OWNER_ROUTER_INTF and
                                'fixed_ips' in p and 'id' in p):
                            req = self.new_action_request(
                                'routers', {'port_id': p['id']}, r['id'],
                                "remove_router_interface")
                            req.get_response(self.ext_api)
                    except n_exc.NeutronException as e:
                        # since this is a cleanup after a UT execution we
                        # just log it and then ignore it. The subsequent router
                        # delete should anyway capture more serious errors
                        LOG.error(_LE('Encountered error during router clean '
                                      'up: '), e)
                # Remove the router, we don't capture any exceptions here as
                # that may hide real bugs
                self._delete('routers', r['id'])
        for rt in self._list('routertypes')['routertypes']:
            # we don't capture any exceptions here as that may hide real bugs
            self._delete('routertypes', rt['id'])


class L3TestRoutertypeExtensionManager(
        l3_router_test_support.TestL3RouterBaseExtensionManager):

    def get_resources(self):
        # most of the resources are added in our super class
        res = super(L3TestRoutertypeExtensionManager, self).get_resources()
        ext_mgr = (device_manager_test_support.
                   TestDeviceManagerExtensionManager())
        for item in ext_mgr.get_resources():
            res.append(item)
        return res


class TestRoutertypeDBPlugin(
    test_db_base_plugin_v2.NeutronDbPluginV2TestCase,
    RoutertypeTestCaseMixin,
    test_db_device_manager.DeviceManagerTestCaseMixin):

    hdm_db.HostingDeviceManagerMixin.path_prefix = "/dev_mgr"
    resource_prefix_map = dict(
        (k, "/dev_mgr")
        for k in ciscodevmgr.RESOURCE_ATTRIBUTE_MAP.keys())

    def setUp(self, core_plugin=None, l3_plugin=None,
              dm_plugin=None, ext_mgr=None):
        if not core_plugin:
            core_plugin = CORE_PLUGIN_KLASS
        if l3_plugin is None:
            l3_plugin = L3_PLUGIN_KLASS
        service_plugins = {'l3_plugin_name': l3_plugin}
        if dm_plugin is not None:
            service_plugins['dm_plugin_name'] = dm_plugin
        cfg.CONF.set_override('api_extensions_path',
                              l3_router_test_support.extensions_path)
        if not ext_mgr:
            ext_mgr = L3TestRoutertypeExtensionManager()
        super(TestRoutertypeDBPlugin, self).setUp(
            plugin=core_plugin, service_plugins=service_plugins,
            ext_mgr=ext_mgr)
        self.l3_plugin = bc.get_plugin(bc.constants.L3)
        # Ensure we use policy definitions from our repo
        cfg.CONF.set_override('policy_file', policy_path, 'oslo_policy')

    def test_create_routertype(self):
        with self.hosting_device_template() as hdt:
            hdt_id = hdt['hosting_device_template']['id']
            attrs = self._get_test_routertype_attr(hdt_id)
            with self.routertype(hdt_id) as rt:
                for k, v in six.iteritems(attrs):
                    self.assertEqual(rt['routertype'][k], v)

    def test_show_routertype(self):
        with self.hosting_device_template() as hdt:
            hdt_id = hdt['hosting_device_template']['id']
            attrs = self._get_test_routertype_attr(hdt_id)
            with self.routertype(hdt_id) as rt:
                req = self.new_show_request('routertypes',
                                            rt['routertype']['id'],
                                            fmt=self.fmt)
                res = self.deserialize(self.fmt,
                                       req.get_response(self.ext_api))
                for k, v in six.iteritems(attrs):
                    self.assertEqual(res['routertype'][k], v)

    def test_show_routertype_non_admin(self):
        with self.hosting_device_template() as hdt:
            hdt_id = hdt['hosting_device_template']['id']
            tenant_id = hdt['hosting_device_template']['tenant_id']
            with self.routertype(hdt_id) as rt:
                rt_id = rt['routertype']['id']
                non_admin_ctx = bc.context.Context('', tenant_id)
                req = self._req('GET', 'routertypes', None, self.fmt, id=rt_id,
                                context=non_admin_ctx)
                res = self.deserialize(self.fmt,
                                       req.get_response(self.ext_api))
                attrs = self._get_non_admin_routertype_attr(hdt_id)
                attrs['id'] = rt_id
                attrs['tenant_id'] = tenant_id
                if bc.NEUTRON_VERSION >= bc.NEUTRON_NEWTON_VERSION:
                    attrs['project_id'] = tenant_id
                self.assertEqual(len(res['routertype']), len(attrs))
                for k, v in six.iteritems(attrs):
                    self.assertEqual(res['routertype'][k], v)

    def test_list_routertypes(self):
        with self.hosting_device_template() as hdt:
            hdt_id = hdt['hosting_device_template']['id']
            with self.routertype(hdt_id, 'rt1') as rt1,\
                    self.routertype(hdt_id, 'rt2') as rt2,\
                    self.routertype(hdt_id, 'rt3') as rt3:
                self._test_list_resources('routertype', [rt1, rt2, rt3],
                                          query_params='template_id=' + hdt_id)

    def test_update_routertype(self):
        with self.hosting_device_template() as hdt:
            hdt_id = hdt['hosting_device_template']['id']
            new_name = 'new_name_1'
            attrs = self._get_test_routertype_attr(hdt_id)
            attrs['name'] = new_name
            with self.routertype(hdt_id) as rt:
                rt_id = rt['routertype']['id']
                data = {'routertype': {'name': new_name}}
                req = self.new_update_request('routertypes', data, rt_id)
                res = self.deserialize(self.fmt,
                                       req.get_response(self.ext_api))
                for k, v in six.iteritems(attrs):
                    self.assertEqual(res['routertype'][k], v)

    def test_delete_routertype(self):
        with self.hosting_device_template() as hdt:
            hdt_id = hdt['hosting_device_template']['id']
            with self.routertype(hdt_id, no_delete=True) as rt:
                rt_id = rt['routertype']['id']
                req = self.new_delete_request('routertypes', rt_id)
                res = req.get_response(self.ext_api)
                self.assertEqual(res.status_int, 204)
                ctx = bc.context.get_admin_context()
                self.assertRaises(routertype.RouterTypeNotFound,
                                  self.l3_plugin.get_routertype, ctx, rt_id)

    def test_routertype_policy(self):
        if (bc.NEUTRON_VERSION < bc.NEUTRON_OCATA_VERSION):
            expect_code = webob.exc.HTTPNotFound.code
        else:
            expect_code = webob.exc.HTTPForbidden.code
        with self.hosting_device_template() as hdt:
            hdt_id = hdt['hosting_device_template']['id']
            tenant_id = hdt['hosting_device_template']['tenant_id']
            with self.routertype(hdt_id) as rt:
                rt_id = rt['routertype']['id']
                non_admin_ctx = bc.context.Context('', tenant_id)
                # create fails
                self._create_routertype(
                    self.fmt, hdt_id, 'fast_routers', 10,
                    webob.exc.HTTPForbidden.code, tenant_id=tenant_id,
                    set_context=True)
                # show succeeds
                self._show('routertypes', rt_id,
                           webob.exc.HTTPOk.code, non_admin_ctx)
                # update fails
                self._update('routertypes', rt_id,
                             {'routertype': {'name': 'new_name'}},
                             webob.exc.HTTPForbidden.code, non_admin_ctx)
                # delete fails
                self._delete('routertypes', rt_id,
                             expect_code, non_admin_ctx)



# Copyright 2015 Cisco Systems, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import eventlet
eventlet.monkey_patch()

from neutron.db import models_v2
from neutron_lib import exceptions as n_exc
from novaclient import exceptions as nova_exc
from oslo_log import log as logging
from sqlalchemy.sql import expression as expr

from networking_cisco._i18n import _
from networking_cisco import backwards_compatibility as bc
from networking_cisco.plugins.cisco.common import cisco_constants
from networking_cisco.plugins.cisco.device_manager import plugging_drivers
from networking_cisco.plugins.cisco.device_manager.plugging_drivers import (
    utils)

LOG = logging.getLogger(__name__)

DELETION_ATTEMPTS = 4
SECONDS_BETWEEN_DELETION_ATTEMPTS = 1


class PortNotUnBoundException(n_exc.InUse):
    message = _("Port: %(port_id)s not unbound yet.")


class VIFHotPlugPluggingDriver(plugging_drivers.PluginSidePluggingDriver,
                               utils.PluggingDriverUtilsMixin):
    """Driver class for service VMs used with Neutron plugins supporting
    VIF hot-plug.
    """

    def __init__(self):
        self._gt_pool = eventlet.GreenPool()

    @property
    def _core_plugin(self):
        return bc.get_plugin()

    @property
    def _dev_mgr(self):
        return bc.get_plugin(cisco_constants.DEVICE_MANAGER)

    def create_hosting_device_resources(self, context, complementary_id,
                                        tenant_id, mgmt_context, max_hosted):
        """Create resources for a hosting device in a plugin specific way."""
        mgmt_port = None
        if mgmt_context and mgmt_context.get('mgmt_nw_id') and tenant_id:
            # Create port for mgmt interface
            p_spec = {'port': {
                'tenant_id': tenant_id,
                'admin_state_up': True,
                'name': 'mgmt',
                'network_id': mgmt_context['mgmt_nw_id'],
                'mac_address': bc.constants.ATTR_NOT_SPECIFIED,
                'fixed_ips': self._mgmt_subnet_spec(context, mgmt_context),
                'device_id': "",
                # Use device_owner attribute to ensure we can query for these
                # ports even before Nova has set device_id attribute.
                'device_owner': complementary_id}}
            try:
                mgmt_port = self._core_plugin.create_port(context, p_spec)
            except n_exc.NeutronException as e:
                LOG.error('Error %s when creating management port. '
                          'Cleaning up.', e)
                self.delete_hosting_device_resources(
                    context, tenant_id, mgmt_port)
                mgmt_port = None
        # We are setting the 'ports' to an empty list as it is expected by
        # the callee: device_handling_db._create_svc_vm_hosting_devices()
        return {'mgmt_port': mgmt_port, 'ports': []}

    def get_hosting_device_resources(self, context, id, complementary_id,
                                     tenant_id, mgmt_nw_id):
        """Returns information about all resources for a hosting device."""
        mgmt_port = None
        # Ports for hosting device may not yet have 'device_id' set to
        # Nova assigned uuid of VM instance. However, those ports will still
        # have 'device_owner' attribute set to complementary_id. Hence, we
        # use both attributes in the query to ensure we find all ports.
        query = context.session.query(models_v2.Port)
        query = query.filter(expr.or_(
            models_v2.Port.device_id == id,
            models_v2.Port.device_owner == complementary_id))
        for port in query:
            if port['network_id'] != mgmt_nw_id:
                raise Exception
            else:
                mgmt_port = port
        return {'mgmt_port': mgmt_port}

    def delete_hosting_device_resources(self, context, tenant_id, mgmt_port,
                                        **kwargs):
        """Deletes resources for a hosting device in a plugin specific way."""

        if mgmt_port is not None:
            try:
                self._cleanup_hosting_port(context, mgmt_port['id'])
            except n_exc.NeutronException as e:
                LOG.error("Unable to delete port:%(port)s after %(tries)d"
                          " attempts due to exception %(exception)s. "
                          "Skipping it", {'port': mgmt_port['id'],
                                          'tries': DELETION_ATTEMPTS,
                                          'exception': str(e)})

    @utils.retry(n_exc.NeutronException, DELETION_ATTEMPTS,
                 SECONDS_BETWEEN_DELETION_ATTEMPTS)
    def _delete_resource_port(self, context, port_id):
        try:
            self._core_plugin.delete_port(context, port_id)
            LOG.debug("Port %s deleted successfully", port_id)
        except n_exc.PortNotFound:
            LOG.warning('Trying to delete port:%s, but port is not found',
                        port_id)

    @utils.retry(PortNotUnBoundException, tries=6)
    def _is_port_unbound(self, context, port_db):
        # Nova will unbind the port asynchronously after the unplug. So we need
        # to expire the port db to fetch the updated info.
        context.session.expire(port_db)
        port_db = self._core_plugin._get_port(context, port_db.id)
        if port_db.device_id == '' and port_db.device_owner == '':
            return True
        else:
            raise PortNotUnBoundException(port_id=port_db.id)

    def _cleanup_hosting_port(self, context, port_id):
        # Needs a bit of time for gateway port to unbind
        eventlet.sleep(2)
        port_db = self._core_plugin._get_port(context, port_id)
        if self._is_port_unbound(context, port_db):
            LOG.debug("Port:%s unbound. Going to delete it", port_db.name)
            self._delete_resource_port(context, port_id)

    def _attach_hosting_port(self, hosting_device_id, hosting_port_id):
        # Give a little bit of time for VM to get out of building state
        eventlet.sleep(10)
        try:
            self.svc_vm_mgr.interface_attach(
                hosting_device_id, hosting_port_id)
            LOG.debug("_attach_hosting_port Setup logical port completed")
        except Exception as e:
            LOG.error("_attach_hosting_port Failed to attach interface:"
                      "%(p_id)s on hosting device:%(hd_id)s due to "
                      "error %(error)s", {'p_id': hosting_port_id,
                                          'hd_id': hosting_device_id,
                                          'error': str(e)})

    def setup_logical_port_connectivity(self, context, port_db,
                                        hosting_device_id):
        """Establishes connectivity for a logical port.

        This is done by hot plugging the interface(VIF) corresponding to the
        port from the VM.
        """
        hosting_port = port_db.hosting_info.hosting_port
        if hosting_port:
            try:
                self._dev_mgr.svc_vm_mgr.interface_attach(hosting_device_id,
                                                          hosting_port.id)
                LOG.debug("Setup logical port completed for port:%s",
                          port_db.id)
            except nova_exc.Conflict as e:
                # VM is still in vm_state building
                LOG.debug("Failed to attach interface - spawn thread "
                          "error %(error)s", {'error': str(e)})
                self._gt_pool.spawn_n(self._attach_hosting_port,
                                      hosting_device_id, hosting_port.id)
            except Exception as e:
                LOG.error("Failed to attach interface mapped to port:"
                          "%(p_id)s on hosting device:%(hd_id)s due to "
                          "error %(error)s", {'p_id': hosting_port.id,
                                              'hd_id': hosting_device_id,
                                              'error': str(e)})

    def teardown_logical_port_connectivity(self, context, port_db,
                                           hosting_device_id):
        """Removes connectivity for a logical port.

        Unplugs the corresponding data interface from the VM.
        """
        if port_db is None or port_db.get('id') is None:
            LOG.warning("Port id is None! Cannot remove port "
                        "from hosting_device:%s", hosting_device_id)
            return
        hosting_port_id = port_db.hosting_info.hosting_port.id
        try:
            self._dev_mgr.svc_vm_mgr.interface_detach(hosting_device_id,
                                                      hosting_port_id)
            self._gt_pool.spawn_n(self._cleanup_hosting_port, context,
                                  hosting_port_id)
            LOG.debug("Teardown logicalport completed for port:%s", port_db.id)

        except Exception as e:
            LOG.error("Failed to detach interface corresponding to port:"
                      "%(p_id)s on hosting device:%(hd_id)s due to "
                      "error %(error)s", {'p_id': hosting_port_id,
                                          'hd_id': hosting_device_id,
                                          'error': str(e)})

    def extend_hosting_port_info(self, context, port_db, hosting_device,
                                 hosting_info):
        """Extends hosting information for a logical port."""
        return

    def allocate_hosting_port(self, context, router_id, port_db, network_type,
                              hosting_device_id):
        """Allocates a hosting port for a logical port.

        We create a hosting port for the router port
        """
        l3admin_tenant_id = self._dev_mgr.l3_tenant_id()
        hostingport_name = 'hostingport_' + port_db['id'][:8]
        p_spec = {'port': {
                  'tenant_id': l3admin_tenant_id,
                  'admin_state_up': True,
                  'name': hostingport_name,
                  'network_id': port_db['network_id'],
                  'mac_address': bc.constants.ATTR_NOT_SPECIFIED,
                  'fixed_ips': [],
                  'device_id': '',
                  'device_owner': '',
                  'port_security_enabled': False}}
        try:
            hosting_port = self._core_plugin.create_port(context, p_spec)
        except n_exc.NeutronException as e:
            LOG.error('Error %s when creating hosting port'
                      'Cleaning up.', e)
            self.delete_hosting_device_resources(
                context, l3admin_tenant_id, hosting_port)
            hosting_port = None
        finally:
            if hosting_port:
                return {'allocated_port_id': hosting_port['id'],
                        'allocated_vlan': None}
            else:
                return None



# Copyright 2014 Cisco Systems, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from datetime import datetime
import eventlet
eventlet.monkey_patch()
import pprint
import sys
import time

from oslo_concurrency import lockutils
from oslo_config import cfg
from oslo_log import log as logging
import oslo_messaging
from oslo_service import loopingcall
from oslo_service import periodic_task
from oslo_service import service
from oslo_utils import importutils

from neutron.agent import rpc as agent_rpc
from neutron.common import config as common_config
from neutron.common import rpc as n_rpc
from neutron.common import topics
from neutron import manager
from neutron import service as neutron_service

from networking_cisco._i18n import _
from networking_cisco import backwards_compatibility as bc
from networking_cisco.backwards_compatibility import neutron_agent_conf
from networking_cisco.plugins.cisco.cfg_agent import device_status
from networking_cisco.plugins.cisco.common import (cisco_constants as
                                                   c_constants)
from networking_cisco.plugins.cisco.device_manager import config


LOG = logging.getLogger(__name__)

# Constants for agent registration.
REGISTRATION_RETRY_DELAY = 2
MAX_REGISTRATION_ATTEMPTS = 30

ISO8601_TIME_FORMAT = bc.constants.ISO8601_TIME_FORMAT


class CiscoDeviceManagementApi(object):
    """Agent side of the device manager RPC API."""

    def __init__(self, topic, host):
        self.host = host
        target = oslo_messaging.Target(topic=topic, version='1.0')
        self.client = n_rpc.get_client(target)

    def report_dead_hosting_devices(self, context, hd_ids=None):
        """Report that a hosting device cannot be contacted (presumed dead).

        :param: context: session context
        :param: hosting_device_ids: list of non-responding hosting devices
        :return: None
        """
        cctxt = self.client.prepare()
        cctxt.cast(context, 'report_non_responding_hosting_devices',
                   host=self.host, hosting_device_ids=hd_ids)

    def report_revived_hosting_devices(self, context, hd_ids=None):
        cctxt = self.client.prepare(version='1.1')
        cctxt.cast(context, 'update_hosting_device_status',
                   host=self.host,
                   status_info={c_constants.HD_ACTIVE: hd_ids})

    def register_for_duty(self, context):
        """Report that a config agent is ready for duty."""
        cctxt = self.client.prepare()
        return cctxt.call(context, 'register_for_duty', host=self.host)

    def get_hosting_devices_for_agent(self, context):
        """Get a list of hosting devices assigned to this agent."""
        cctxt = self.client.prepare()
        return cctxt.call(context,
                          'get_hosting_devices_for_agent',
                          host=self.host)

OPTS = [
    cfg.IntOpt('rpc_loop_interval', default=10,
               help=_("Interval when the process_services() loop "
                      "executes in seconds. This is when the config agent "
                      "lets each service helper to process its neutron "
                      "resources.")),
    cfg.StrOpt('routing_svc_helper_class',
               default='networking_cisco.plugins.cisco.cfg_agent.'
                       'service_helpers.routing_svc_helper.'
                       'RoutingServiceHelper',
               help=_("Path of the routing service helper class.")),
    cfg.BoolOpt('enable_heartbeat',
                default=True,
                help=_("If enabled, the agent will maintain a heartbeat "
                       "against its hosting-devices. If a device dies "
                       "and recovers, the agent will then trigger a "
                       "configuration resync.")),
    cfg.IntOpt('heartbeat_interval', default=5,
               help=_("Interval in seconds when the config agent runs the "
                      "backlog / hosting-device heart beat task.")),
    cfg.IntOpt('max_device_sync_attempts', default=6,
               help=_("Maximum number of attempts for a device sync.")),
    cfg.IntOpt('max_device_sync_batch_size', default=64,
               help=_("The largest number of routers to fetch in one "
                      "RPC call.")),
    cfg.IntOpt('keepalive_interval', default=10,
               help=_("Interval in seconds when the config agent sents a "
                      "timestamp to the plugin to say that it is alive.")),
    cfg.IntOpt('report_iteration', default=6,
               help=_("The iteration where the config agent sends a full "
                      "status report to the plugin.  The default is every "
                      "6th iteration of the keep alive interval. This "
                      "means with default value of keepalive_interval "
                      "(10sec), a full report is sent once every "
                      "6*10 = 60 seconds.")),
]

cfg.CONF.register_opts(OPTS, "cfg_agent")


class CiscoCfgAgent(manager.Manager):
    """Cisco Cfg Agent.

    This class defines a generic configuration agent for cisco devices which
    implement network services in the cloud backend. It is based on the
    (reference) l3-agent, but has been enhanced to support multiple services
    in addition to routing.

    The agent acts like as a container for services and does not do any
    service specific processing or configuration itself.
    All service specific processing is delegated to service helpers which
    the agent loads. Thus routing specific updates are processed by the
    routing service helper, firewall by firewall helper etc.
    A further layer of abstraction is implemented by using device drivers for
    encapsulating all configuration operations of a service on a device.
    Device drivers are specific to a particular device/service VM eg: ASR1k.

    The main entry points in this class are the `process_services()` and
    `_backlog_task()` .
    """
    target = oslo_messaging.Target(version='1.1')

    def __init__(self, host, conf=None):
        self.conf = conf or cfg.CONF
        self._credentials = (
            config.obtain_hosting_device_credentials_from_config())
        self._dev_status = device_status.DeviceStatus()
        self._dev_status.enable_heartbeat = (
            self.conf.cfg_agent.enable_heartbeat)
        self.context = bc.context.get_admin_context_without_session()

        self._initialize_rpc(host)
        self._initialize_service_helpers(host)
        self._start_periodic_tasks()
        super(CiscoCfgAgent, self).__init__(host=self.conf.host)

    def _initialize_rpc(self, host):
        self.devmgr_rpc = CiscoDeviceManagementApi(
            c_constants.DEVICE_MANAGER_PLUGIN, host)

    def _initialize_service_helpers(self, host):
        svc_helper_class = self.conf.cfg_agent.routing_svc_helper_class
        try:
            self.routing_service_helper = importutils.import_object(
                svc_helper_class, host, self.conf, self)
        except ImportError as e:
            LOG.warning("Error in loading routing service helper. Class "
                        "specified is %(class)s. Reason:%(reason)s",
                        {'class': self.conf.cfg_agent.routing_svc_helper_class,
                         'reason': e})
            self.routing_service_helper = None

    def _start_periodic_tasks(self):
        self.loop = loopingcall.FixedIntervalLoopingCall(self.process_services)
        self.loop.start(interval=self.conf.cfg_agent.rpc_loop_interval)

    def after_start(self):
        hds = self.get_assigned_hosting_devices()
        if hds is None:
            LOG.warning("Unable to fetch hosting devices to monitor. Certain "
                        "hosting devices in DEAD state may not be revived. "
                        "Restart config agent to trigger new attempts.")
        else:
            LOG.info("To manage hosting devices %s", hds)
            self._dev_status.backlog_hosting_devices(hds)
        LOG.info("Cisco cfg agent started")

    def get_routing_service_helper(self):
        return self.routing_service_helper

    def get_hosting_device_password(self, credentials_id):
        creds = self._credentials.get(credentials_id)
        return creds['password'] if creds else None

    ## Periodic tasks ##
    @periodic_task.periodic_task(spacing=cfg.CONF.cfg_agent.heartbeat_interval)
    def _backlog_task(self, context):
        """Process backlogged devices."""
        LOG.debug("Processing backlog.")
        self._process_backlogged_hosting_devices(context)

    ## Main orchestrator ##
    @lockutils.synchronized('cisco-cfg-agent', 'neutron-')
    def process_services(self, device_ids=None, removed_devices_info=None):
        """Process services managed by this config agent.

        This method is invoked by any of three scenarios.

        1. Invoked by a periodic task running every `RPC_LOOP_INTERVAL`
        seconds. This is the most common scenario.
        In this mode, the method is called without any arguments.

        2. Called by the `_process_backlogged_hosting_devices()` as part of
        the backlog processing task. In this mode, a list of device_ids
        are passed as arguments. These are the list of backlogged
        hosting devices that are now reachable and we want to sync services
        on them.

        3. Called by the `hosting_devices_removed()` method. This is when
        the config agent has received a notification from the plugin that
        some hosting devices are going to be removed. The payload contains
        the details of the hosting devices and the associated neutron
        resources on them which should be processed and removed.

        To avoid race conditions with these scenarios, this function is
        protected by a lock.

        This method goes on to invoke `process_service()` on the
        different service helpers.

        :param device_ids: List of devices that are now available and needs to
            be processed
        :param removed_devices_info: Info about the hosting devices which are
            going to be removed and details of the resources hosted on them.
            Expected Format::

                {
                 'hosting_data': {'hd_id1': {'routers': [id1, id2, ...]},
                                  'hd_id2': {'routers': [id3, id4, ...]}, ...},
                 'deconfigure': True/False
                }

        :returns: None

        """
        LOG.debug("Processing services started")
        # Now we process only routing service, additional services will be
        # added in future
        if self.routing_service_helper:
            self.routing_service_helper.process_service(device_ids,
                                                        removed_devices_info)
        else:
            LOG.warning("No routing service helper loaded")
        LOG.debug("Processing services completed")

    def _process_backlogged_hosting_devices(self, context):
        """Process currently backlogged devices.

        Go through the currently backlogged devices and process them.
        For devices which are now reachable (compared to last time), we call
        `process_services()` passing the now reachable device's id.
        For devices which have passed the `hosting_device_dead_timeout` and
        hence presumed dead, execute a RPC to the plugin informing that.

        heartbeat revision
        res['reachable'] - hosting device went from Unknown to Active state
                           process_services(...)
        res['revived']   - hosting device went from Dead to Active
                           inform device manager that the hosting
                           device is now responsive
        res['dead']      - hosting device went from Unknown to Dead
                           inform device manager that the hosting
                           device is non-responding

        As additional note for the revived case:
            Although the plugin was notified, there may be some lag
            before the plugin actually can reschedule it's backlogged routers.

            If process_services(device_ids...) isn't successful initially,
            subsequent device syncs will be attempted until
            MAX_DEVICE_SYNC_ATTEMPTS occurs.  Main process_service task
            will resume if sync_devices is populated.

        :param context: RPC context
        :return: None
        """
        driver_mgr = self.get_routing_service_helper().driver_manager
        res = self._dev_status.check_backlogged_hosting_devices(driver_mgr)
        if res['reachable']:
            self.process_services(device_ids=res['reachable'])
        if res['revived']:
            LOG.debug("Reporting revived hosting devices: %s " %
                      res['revived'])
            # trigger a sync only on the revived hosting-devices
            if self.conf.cfg_agent.enable_heartbeat is True:
                self.devmgr_rpc.report_revived_hosting_devices(
                    context, hd_ids=res['revived'])
                self.process_services(device_ids=res['revived'])
        if res['dead']:
            LOG.debug("Reporting dead hosting devices: %s", res['dead'])
            self.devmgr_rpc.report_dead_hosting_devices(context,
                                                        hd_ids=res['dead'])

    def agent_updated(self, context, payload):
        """Deal with agent updated RPC message."""
        try:
            if payload['admin_state_up']:
                #TODO(hareeshp): implement agent updated handling
                pass
        except KeyError as e:
            LOG.error("Invalid payload format for received RPC message "
                      "`agent_updated`. Error is %(error)s. Payload is "
                      "%(payload)s", {'error': e, 'payload': payload})

    def hosting_devices_assigned_to_cfg_agent(self, context, payload):
        """Deal with hosting devices assigned to this config agent."""
        LOG.debug("Got hosting device assigned, payload: %s" % payload)
        try:
            if payload['hosting_device_ids']:
                #TODO(hareeshp): implement assignment of hosting devices
                self.routing_service_helper.fullsync = True
        except KeyError as e:
            LOG.error("Invalid payload format for received RPC message "
                      "`hosting_devices_assigned_to_cfg_agent`. Error is "
                      "%(error)s. Payload is %(payload)s",
                      {'error': e, 'payload': payload})

    def hosting_devices_unassigned_from_cfg_agent(self, context, payload):
        """Deal with hosting devices unassigned from this config agent."""
        try:
            if payload['hosting_device_ids']:
                #TODO(hareeshp): implement unassignment of hosting devices
                pass
        except KeyError as e:
            LOG.error("Invalid payload format for received RPC message "
                      "`hosting_devices_unassigned_from_cfg_agent`. Error "
                      "is %(error)s. Payload is %(payload)s",
                      {'error': e, 'payload': payload})

    def hosting_devices_removed(self, context, payload):
        """Deal with hosting device removed RPC message."""
        try:
            if payload['hosting_data']:
                if payload['hosting_data'].keys():
                    self.process_services(removed_devices_info=payload)
        except KeyError as e:
            LOG.error("Invalid payload format for received RPC message "
                      "`hosting_devices_removed`. Error is %(error)s. Payload "
                      "is %(payload)s", {'error': e, 'payload': payload})

    def get_assigned_hosting_devices(self, max_retry_attempts=3):
        context = bc.context.get_admin_context_without_session()
        for attempts in range(max_retry_attempts):
            try:
                return self.devmgr_rpc.get_hosting_devices_for_agent(context)
            except Exception:
                LOG.warning("Failed attempt to fetch hosting devices to "
                            "manage. Retrying in %0.2f second.",
                            REGISTRATION_RETRY_DELAY)
                time.sleep(REGISTRATION_RETRY_DELAY)
        return None

    def get_hosting_device_configuration(self, context, payload):
        LOG.debug('Processing request to fetching running config')
        hd_id = payload['hosting_device_id']
        svc_helper = self.routing_service_helper
        if hd_id and svc_helper:
            LOG.debug('Fetching running config for %s' % hd_id)
            drv = svc_helper.driver_manager.get_driver_for_hosting_device(
                hd_id)
            rc = drv.get_configuration()
            if rc:
                LOG.debug('Fetched %(chars)d characters long running config '
                          'for %(hd_id)s' % {'chars': len(rc), 'hd_id': hd_id})
                return rc
        LOG.debug('Unable to get running config')
        return


class CiscoCfgAgentWithStateReport(CiscoCfgAgent):

    def __init__(self, host, conf=None):
        self.state_rpc = agent_rpc.PluginReportStateAPI(topics.PLUGIN)
        self.agent_state = {
            'binary': 'neutron-cisco-cfg-agent',
            'host': host,
            'topic': c_constants.CFG_AGENT,
            'configurations': {},
            'start_flag': True,
            'agent_type': c_constants.AGENT_TYPE_CFG}
        self.use_call = True
        self._initialize_rpc(host)
        self._agent_registration()
        super(CiscoCfgAgentWithStateReport, self).__init__(host=host,
                                                           conf=conf)

        self.report_iteration = self.conf.cfg_agent.report_iteration
        keepalive_interval = self.conf.cfg_agent.keepalive_interval
        if keepalive_interval:
            self.heartbeat = loopingcall.FixedIntervalLoopingCall(
                self._report_state)
            self.heartbeat.start(interval=keepalive_interval)
            self.keepalive_iteration = 0

    def _agent_registration(self):
        """Register this agent with the server.

        This method registers the cfg agent with the neutron server so hosting
        devices can be assigned to it. In case the server is not ready to
        accept registration (it sends a False) then we retry registration
        for `MAX_REGISTRATION_ATTEMPTS` with a delay of
        `REGISTRATION_RETRY_DELAY`. If there is no server response or a
        failure to register after the required number of attempts,
        the agent stops itself.
        """
        for attempts in range(MAX_REGISTRATION_ATTEMPTS):
            context = bc.context.get_admin_context_without_session()
            self.send_agent_report(self.agent_state, context)
            try:
                res = self.devmgr_rpc.register_for_duty(context)
            except Exception:
                res = False
                LOG.warning("[Agent registration] Rpc exception. Neutron "
                            "may not be available or busy. Retrying "
                            "in %0.2f seconds ", REGISTRATION_RETRY_DELAY)
            if res is True:
                LOG.info("[Agent registration] Agent successfully registered")
                return
            elif res is False:
                LOG.warning("[Agent registration] Neutron server said "
                            "that device manager was not ready. Retrying "
                            "in %0.2f seconds ", REGISTRATION_RETRY_DELAY)
                time.sleep(REGISTRATION_RETRY_DELAY)
            elif res is None:
                LOG.error("[Agent registration] Neutron server said that "
                          "no device manager was found. Cannot continue. "
                          "Exiting!")
                raise SystemExit(_("Cfg Agent exiting"))
        LOG.error("[Agent registration] %d unsuccessful registration "
                  "attempts. Exiting!", MAX_REGISTRATION_ATTEMPTS)
        raise SystemExit(_("Cfg Agent exiting"))

    def _report_state(self):
        """Report state to the plugin.

        This task run every `keepalive_interval` period.
        Collects, creates and sends a summary of the services currently
        managed by this agent. Data is collected from the service helper(s).
        Refer the `configurations` dict for the parameters reported.
        :return: None
        """
        LOG.debug("Report state task started")
        self.keepalive_iteration += 1
        if self.keepalive_iteration == self.report_iteration:
            self._prepare_full_report_data()
            self.keepalive_iteration = 0
            LOG.debug("State report: %s", pprint.pformat(self.agent_state))
        else:
            self.agent_state.pop('configurations', None)
            self.agent_state['local_time'] = datetime.now().strftime(
                ISO8601_TIME_FORMAT)
            LOG.debug("State report: %s", self.agent_state)
        self.send_agent_report(self.agent_state, self.context)

    def _prepare_full_report_data(self):
        configurations = {}
        service_agents = []
        self.agent_state['configurations'] = configurations
        if self.routing_service_helper:
            service_agents.append(c_constants.AGENT_TYPE_L3_CFG)
            configurations = self.routing_service_helper.collect_state(
                self.agent_state['configurations'])
        non_responding = self._dev_status.get_backlogged_hosting_devices_info()
        monitored_hosting_devices = (self._dev_status.
                                     get_monitored_hosting_devices_info())
        configurations['non_responding_hosting_devices'] = non_responding
        configurations['monitored_hosting_devices'] = monitored_hosting_devices
        configurations['service_agents'] = service_agents
        self.agent_state['configurations'] = configurations
        self.agent_state['local_time'] = datetime.now().strftime(
            ISO8601_TIME_FORMAT)

    def send_agent_report(self, report, context):
        """Send the agent report via RPC."""
        try:
            self.state_rpc.report_state(context, report, self.use_call)
            report.pop('start_flag', None)
            self.use_call = False
            LOG.debug("Send agent report successfully completed")
        except AttributeError:
            # This means the server does not support report_state
            LOG.warning("Neutron server does not support state report. "
                        "State report for this agent will be disabled.")
            self.heartbeat.stop()
            return
        except Exception:
            LOG.warning("Failed sending agent report!")


def _mock_stuff():
    import mock

    targets = ['networking_cisco.plugins.cisco.cfg_agent.device_drivers.'
               'iosxe.iosxe_routing_driver.manager']
    ncc_patchers = []
    ncclient_mock = mock.MagicMock()
    ok_xml_obj = mock.MagicMock()
    ok_xml_obj.xml = "<ok />"
    ncclient_mock.connect.return_value.edit_config.return_value = ok_xml_obj
    for target in targets:
        patcher = mock.patch(target, ncclient_mock)
        patcher.start()
        ncc_patchers.append(patcher)

    targets = ['networking_cisco.plugins.cisco.cfg_agent.device_drivers.'
               'iosxe.iosxe_routing_driver.IosXeRoutingDriver.'
               '_get_running_config',
               'networking_cisco.plugins.cisco.cfg_agent.device_drivers.asr1k.'
               'asr1k_cfg_syncer.ConfigSyncer.get_running_config']
    fake_running_config = ("interface GigabitEthernet1\n"
                           "ip address 10.0.0.10 255.255.255.255\n"
                           "ip route 0.0.0.0 0.0.0.0 GigabitEthernet1 "
                           "10.0.0.1")
    g_r_c_patchers = []
    g_r_c_mock = mock.MagicMock(return_value=fake_running_config)
    for target in targets:
        patcher = mock.patch(target, g_r_c_mock)
        patcher.start()
        g_r_c_patchers.append(patcher)

    is_pingable_mock = mock.MagicMock(return_value=True)
    pingable_patcher = mock.patch(
        'networking_cisco.plugins.cisco.cfg_agent.device_status._is_pingable',
        is_pingable_mock)
    pingable_patcher.start()


# NOTE(bobmel): call mock_ncclient() in main() of cfg_agent.py to run config
# agent with a fake ncclient. That mocked mode of running the config agent is
# useful for end-2-end-like debugging without actual backend hosting devices.
def mock_ncclient():
    import mock
    from ncclient.transport import errors as ncc_exc
    from networking_cisco.plugins.cisco.common import (cisco_ios_xe_simulator
        as cisco_ios_xe)
    import os

    def _fake_connect(host, port, username, password, device_params, timeout):
        sim = cisco_ios_xe.CiscoIOSXESimulator(
            '', host, "255.255.255.0", port, username, password,
            device_params, "GigabitEthernet0", timeout)
        connection_mock = mock.MagicMock()
        connection_mock.edit_config.side_effect = _get_fake_edit_config(sim)
        connection_mock.get_config.side_effect = (
            _get_fake_get_config(sim))
        return connection_mock

    def _get_fake_edit_config(simulator):

        def edit_config(config, format='xml', target='candidate',
                        default_operation=None, test_option=None,
                        error_option=None):
            if not _fake_is_pingable(rc_emul.host_ip):
                raise ncc_exc.SSHError('SSH error')
            rc_emul.edit_config(config)
            print(rc_emul.get_config())
            return ok_xml_obj

        ok_xml_obj = mock.MagicMock()
        ok_xml_obj.xml = "<ok />"
        rc_emul = simulator
        return edit_config

    def _get_fake_get_config(simulator):

        def get_running_config(source):
            if not _fake_is_pingable(rc_emul.host_ip):
                raise ncc_exc.SSHError('SSH error')
            head = ('<?xml version="1.0" encoding="UTF-8"?><rpc-reply '
                    'message-id="urn:uuid:ec8bab72-a500-11e5-a92f'
                    '-74a2e6d55908" xmlns="urn:ietf:params:xml:ns:netconf:'
                    'base:1.0"><data><cli-config-data-block>!')
            tail = '</cli-config-data-block></data></rpc-reply>'
            raw_rc = rc_emul.get_config()
            return cisco_ios_xe.FakeRunningConfig(head + raw_rc + tail)

        rc_emul = simulator
        return get_running_config

    def _fake_is_pingable(ip):
        # if a file with a certain name (derived from the 'ip' argument):
        #
        #     /opt/stack/data/neutron/DEAD__10_0_5_8       (ip = 10.0.5.8)
        #
        # exists then the (faked) hosting device with that IP address
        # will appear to NOT respond to pings
        path = '/opt/stack/data/neutron'
        indicator_filename = path + '/DEAD_' + str(ip).replace('.', '_')
        return not os.path.isfile(indicator_filename)

    def _fake_can_connect(ip, port):
        # if a file with a certain name (derived from the 'ip' and
        # 'port's arguments):
        #
        #     /opt/stack/data/neutron/DEAD__10_0_5_8__22       (ip = 10.0.5.8
        #                                                       port = 22)
        #
        # exists then the (faked) hosting device with that IP address
        # will appear NOT to have that port open
        path = '/opt/stack/data/neutron'
        indicator_filename = (path + '/DEAD__' + str(ip).replace('.', '_') +
                              '__' + str(port))
        return not os.path.isfile(indicator_filename)

    targets = ['networking_cisco.plugins.cisco.cfg_agent.device_drivers.'
               'iosxe.iosxe_routing_driver.manager']
    ncc_patchers = []
    ncclient_mgr_mock = mock.MagicMock()
    ncclient_mgr_mock.connect.side_effect = _fake_connect

    for target in targets:
        patcher = mock.patch(target, ncclient_mgr_mock)
        patcher.start()
        ncc_patchers.append(patcher)

    is_pingable_mock = mock.MagicMock()
    is_pingable_mock.side_effect = _fake_is_pingable
    pingable_patcher = mock.patch(
        'networking_cisco.plugins.cisco.cfg_agent.device_status._is_pingable',
        is_pingable_mock)
    pingable_patcher.start()

    can_connect_mock = mock.MagicMock(side_effect=_fake_can_connect)
    can_connect_patcher = mock.patch(
        'networking_cisco.plugins.cisco.cfg_agent.device_status._can_connect',
        can_connect_mock)
    can_connect_patcher.start()


def main(manager='networking_cisco.plugins.cisco.cfg_agent.'
                 'cfg_agent.CiscoCfgAgentWithStateReport'):
    # NOTE(bobmel): call _mock_stuff() to run config agent with fake ncclient
    # This mocked mode of running the config agent is useful for end-2-end-like
    # debugging without actual backend hosting devices.
    #_mock_stuff()
    #mock_ncclient()
    conf = cfg.CONF
    conf.register_opts(OPTS, "cfg_agent")
    bc.config.register_agent_state_opts_helper(conf)
    bc.config.register_root_helper(conf)
    conf.register_opts(neutron_agent_conf.INTERFACE_OPTS)
    conf.register_opts(neutron_agent_conf.EXTERNAL_PROCESS_OPTS)
    common_config.init(sys.argv[1:])
    conf(project='neutron')
    bc.config.setup_logging()
    server = neutron_service.Service.create(
        binary='neutron-cisco-cfg-agent',
        topic=c_constants.CFG_AGENT,
        report_interval=cfg.CONF.AGENT.report_interval,
        manager=manager)
    service.launch(cfg.CONF, server).wait()



# Copyright 2014 Cisco Systems, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from oslo_log import log as logging
import oslo_messaging

from neutron.common import rpc as n_rpc
from neutron.common import topics
from neutron.common import utils

from networking_cisco import backwards_compatibility as bc
from networking_cisco.plugins.cisco.common import cisco_constants
from networking_cisco.plugins.cisco.extensions import ciscocfgagentscheduler

LOG = logging.getLogger(__name__)


L3AGENT_SCHED = bc.constants.L3_AGENT_SCHEDULER_EXT_ALIAS
CFGAGENT_SCHED = ciscocfgagentscheduler.CFG_AGENT_SCHEDULER_ALIAS
CFG_AGENT_L3_ROUTING = cisco_constants.CFG_AGENT_L3_ROUTING


class L3RouterCfgAgentNotifyAPI(object):
    """API for plugin to notify Cisco cfg agent."""

    def __init__(self, l3plugin, topic=CFG_AGENT_L3_ROUTING):
        self._l3plugin = l3plugin
        target = oslo_messaging.Target(topic=topic, version='1.0')
        self.client = n_rpc.get_client(target)

    def _agent_notification_bulk(self, context, method, routers,
                                 hosting_device, operation):
        """Notify the Cisco cfg agent handling a particular hosting_device.

        A single notification can contain multiple routers.
        """
        admin_context = context.is_admin and context or context.elevated()
        dmplugin = bc.get_plugin(cisco_constants.DEVICE_MANAGER)
        if (hosting_device is not None and utils.is_extension_supported(
                dmplugin, CFGAGENT_SCHED)):
            agents = dmplugin.get_cfg_agents_for_hosting_devices(
                admin_context, [hosting_device['id']], admin_state_up=True,
                schedule=True)
            if agents:
                agent = agents[0]
                LOG.debug('Notify %(agent_type)s at %(topic)s.%(host)s the '
                          'message %(method)s [BULK]',
                          {'agent_type': agent.agent_type,
                           'topic': CFG_AGENT_L3_ROUTING,
                           'host': agent.host,
                           'method': method})
                cctxt = self.client.prepare(server=agent.host,
                                            version='1.1')
                cctxt.cast(context, method, routers=routers)

    def _agent_notification(self, context, method, routers, operation,
                            shuffle_agents):
        """Notify individual Cisco cfg agents."""
        admin_context = context.is_admin and context or context.elevated()
        dmplugin = bc.get_plugin(cisco_constants.DEVICE_MANAGER)
        for router in routers:
            if (router['hosting_device'] is not None and
                    utils.is_extension_supported(dmplugin, CFGAGENT_SCHED)):
                agents = dmplugin.get_cfg_agents_for_hosting_devices(
                    admin_context, [router['hosting_device']['id']],
                    admin_state_up=True, schedule=True)
            else:
                continue
            for agent in agents:
                LOG.debug('Notify %(agent_type)s at %(topic)s.%(host)s the '
                          'message %(method)s',
                          {'agent_type': agent.agent_type,
                           'topic': CFG_AGENT_L3_ROUTING,
                           'host': agent.host,
                           'method': method})
                cctxt = self.client.prepare(server=agent.host)
                cctxt.cast(context, method, routers=[router['id']])

    def _notification(self, context, method, routers, operation,
                      shuffle_agents):
        """Notify all or individual Cisco cfg agents."""
        if utils.is_extension_supported(self._l3plugin, L3AGENT_SCHED):
            adm_context = (context.is_admin and context or context.elevated())
            # This is where hosting device gets scheduled to Cisco cfg agent
            self._l3plugin.schedule_routers(adm_context, routers)
            self._agent_notification(
                context, method, routers, operation, shuffle_agents)
        else:
            cctxt = self.client.prepare(topics=topics.L3_AGENT, fanout=True)
            cctxt.cast(context, method, routers=[r['id'] for r in routers])

    def router_deleted(self, context, router):
        """Notifies cfg agents about a deleted router."""
        self._agent_notification(context, 'router_deleted', [router], None,
                                 False)

    def routers_updated(self, context, routers, operation=None, data=None,
                        shuffle_agents=False):
        """Notify cfg agents about configuration changes to routers.

        This includes operations performed on the router like when a
        router interface is added or removed.
        """
        if routers:
            self._notification(context, 'routers_updated', routers, operation,
                               shuffle_agents)

    def router_removed_from_hosting_device(self, context, router):
        """Notify cfg agent about router removed from hosting device."""
        self._notification(context, 'router_removed_from_hosting_device',
                           [router], operation=None, shuffle_agents=False)

    def router_added_to_hosting_device(self, context, router):
        """Notify cfg agent about router added to hosting device."""
        self._notification(context, 'router_added_to_hosting_device',
                           [router], operation=None, shuffle_agents=False)

    def routers_removed_from_hosting_device(self, context, router_ids,
                                            hosting_device):
        """Notify cfg agent that routers have been removed from hosting device.
        @param: context - information about tenant, user etc
        @param: router-ids - list of ids
        @param: hosting_device - device hosting the routers
        """
        self._agent_notification_bulk(
            context, 'router_removed_from_hosting_device', router_ids,
            hosting_device, operation=None)



#Copyright (c) 2017 Cisco Systems, Inc.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
VXLAN Test Class using RESTAPI Driver to test Cisco Nexus platforms.

These Classes are based on the original ssh VXLAN event driver
so same tests occur with same configuration.  What's different
between the tests is the resulting driver output which is what
the tests in this class presents to its parent class.

You will notice in this file there are test methods which
are skipped by using 'pass'.  This is because these tests
apply to ssh only OR because rerunning the test would be
redundant.
"""


from oslo_config import cfg


from networking_cisco.ml2_drivers.nexus import (
    nexus_restapi_snippets as snipp)
from networking_cisco.tests.unit.ml2_drivers.nexus import (
    test_cisco_nexus_base as base)
from networking_cisco.tests.unit.ml2_drivers.nexus import (
    test_cisco_nexus_events_vxlan)


class TestCiscoNexusRestVxlanResults(base.TestCiscoNexusBaseResults):

    """Unit tests driver results for Cisco ML2 Nexus."""

    test_results = {

        # The following contains desired Nexus output for
        # some basic vxlan config.
        'add_port_driver_result': [
            [(snipp.PATH_VNI_UPDATE % ('1', '70000')),
             base.NEXUS_IP_ADDRESS_1,
             (snipp.BODY_VNI_UPDATE % (
                 '70000', '70000', '70000',
                 base.MCAST_GROUP)),
             base.POST],
            [snipp.PATH_ALL,
             base.NEXUS_IP_ADDRESS_1,
             (snipp.BODY_VXLAN_ADD % (267, 70000)),
             base.POST],
            [(snipp.PATH_IF % 'phys-[eth1/10]'),
             base.NEXUS_IP_ADDRESS_1,
             (snipp.BODY_TRUNKVLAN % ('l1PhysIf', '', '+267')),
             base.POST]
        ],

        'delete_port_driver_result': [
            [(snipp.PATH_VNI_UPDATE % ('1', '70000')),
             base.NEXUS_IP_ADDRESS_1,
             '',
             base.DELETE],
            [(snipp.PATH_IF % 'phys-[eth1/10]'),
             base.NEXUS_IP_ADDRESS_1,
             (snipp.BODY_TRUNKVLAN % ('l1PhysIf', '', '-267')),
             base.POST],
            [(snipp.PATH_VLAN % '267'),
             base.NEXUS_IP_ADDRESS_1,
             '',
             base.DELETE]
        ],


        'add_port2_driver_result': [
            [snipp.PATH_ALL,
             base.NEXUS_IP_ADDRESS_1,
             (snipp.BODY_VXLAN_ADD % (267, 70000)),
             base.POST],
            [(snipp.PATH_IF % 'phys-[eth1/20]'),
             base.NEXUS_IP_ADDRESS_1,
             (snipp.BODY_TRUNKVLAN % ('l1PhysIf', '', '+267')),
             base.POST]
        ],

        'delete_port2_driver_result': [
            [(snipp.PATH_IF % 'phys-[eth1/20]'),
             base.NEXUS_IP_ADDRESS_1,
             (snipp.BODY_TRUNKVLAN % ('l1PhysIf', '', '-267')),
             base.POST],
        ],

        'add_port_driver_result3': [
            [(snipp.PATH_VNI_UPDATE % ('1', '70000')),
             base.NEXUS_IP_ADDRESS_6,
             (snipp.BODY_VNI_UPDATE % (
                 '70000', '70000', '70000',
                 base.MCAST_GROUP)),
             base.POST],
            [(snipp.PATH_VNI_UPDATE % ('1', '70000')),
             base.NEXUS_IP_ADDRESS_7,
             (snipp.BODY_VNI_UPDATE % (
                 '70000', '70000', '70000',
                 base.MCAST_GROUP)),
             base.POST],
            [snipp.PATH_ALL,
             base.NEXUS_IP_ADDRESS_6,
             (snipp.BODY_VXLAN_ADD % (267, 70000)),
             base.POST],
            [(snipp.PATH_IF % 'phys-[eth1/10]'),
             base.NEXUS_IP_ADDRESS_6,
             (snipp.BODY_TRUNKVLAN % ('l1PhysIf', '', '+267')),
             base.POST],
            [snipp.PATH_ALL,
             base.NEXUS_IP_ADDRESS_7,
             (snipp.BODY_VXLAN_ADD % (267, 70000)),
             base.POST],
            [(snipp.PATH_IF % 'phys-[eth1/2]'),
             base.NEXUS_IP_ADDRESS_7,
             (snipp.BODY_TRUNKVLAN % ('l1PhysIf', '', '+267')),
             base.POST],
            [snipp.PATH_ALL,
             base.NEXUS_IP_ADDRESS_7,
             (snipp.BODY_VXLAN_ADD % (267, 70000)),
             base.POST],
            [(snipp.PATH_IF % 'phys-[eth1/3]'),
             base.NEXUS_IP_ADDRESS_7,
             (snipp.BODY_TRUNKVLAN % ('l1PhysIf', '', '+267')),
             base.POST]
        ],

        'delete_port_driver_result3': [
            [(snipp.PATH_VNI_UPDATE % ('1', '70000')),
             base.NEXUS_IP_ADDRESS_6,
             '',
             base.DELETE],
            [(snipp.PATH_VNI_UPDATE % ('1', '70000')),
             base.NEXUS_IP_ADDRESS_7,
             '',
             base.DELETE],
            [(snipp.PATH_IF % 'phys-[eth1/10]'),
             base.NEXUS_IP_ADDRESS_6,
             (snipp.BODY_TRUNKVLAN % ('l1PhysIf', '', '-267')),
             base.POST],
            [(snipp.PATH_VLAN % '267'),
             base.NEXUS_IP_ADDRESS_6,
             '',
             base.DELETE],
            [(snipp.PATH_IF % 'phys-[eth1/2]'),
             base.NEXUS_IP_ADDRESS_7,
             (snipp.BODY_TRUNKVLAN % ('l1PhysIf', '', '-267')),
             base.POST],
            [(snipp.PATH_VLAN % '267'),
             base.NEXUS_IP_ADDRESS_7,
             '',
             base.DELETE],
            [(snipp.PATH_IF % 'phys-[eth1/3]'),
             base.NEXUS_IP_ADDRESS_7,
             (snipp.BODY_TRUNKVLAN % ('l1PhysIf', '', '-267')),
             base.POST],
        ],

        'add_port_driver_result2': [
            [(snipp.PATH_VNI_UPDATE % ('1', '70001')),
             base.NEXUS_IP_ADDRESS_8,
             (snipp.BODY_VNI_UPDATE % (
                 '70001', '70001', '70001',
                 base.MCAST_GROUP)),
             base.POST],
            [snipp.PATH_ALL,
             base.NEXUS_IP_ADDRESS_8,
             (snipp.BODY_VXLAN_ADD % (265, 70001)),
             base.POST],
            [(snipp.PATH_IF % 'phys-[eth1/20]'),
             base.NEXUS_IP_ADDRESS_8,
             (snipp.BODY_TRUNKVLAN % ('l1PhysIf', '', '+265')),
             base.POST]
        ],

        'delete_port_driver_result2': [
            [(snipp.PATH_VNI_UPDATE % ('1', '70001')),
             base.NEXUS_IP_ADDRESS_8,
             '',
             base.DELETE],
            [(snipp.PATH_IF % 'phys-[eth1/20]'),
             base.NEXUS_IP_ADDRESS_8,
             (snipp.BODY_TRUNKVLAN % ('l1PhysIf', '', '-265')),
             base.POST],
            [(snipp.PATH_VLAN % '265'),
             base.NEXUS_IP_ADDRESS_8,
             '',
             base.DELETE]
        ],
        'add_port_driver_result4': [
            [(snipp.PATH_VNI_UPDATE % ('1', '70000')),
             base.NEXUS_IP_ADDRESS_8,
             (snipp.BODY_VNI_UPDATE % (
                 '70000', '70000', '70000',
                 base.MCAST_GROUP)),
             base.POST],
            [snipp.PATH_ALL,
             base.NEXUS_IP_ADDRESS_8,
             (snipp.BODY_VXLAN_ADD % (267, 70000)),
             base.POST],
            [(snipp.PATH_IF % 'phys-[eth1/10]'),
             base.NEXUS_IP_ADDRESS_8,
             (snipp.BODY_TRUNKVLAN % ('l1PhysIf', '', '+267')),
             base.POST]
        ],

        'delete_port_driver_result4': [
            [(snipp.PATH_VNI_UPDATE % ('1', '70000')),
             base.NEXUS_IP_ADDRESS_8,
             '',
             base.DELETE],
            [(snipp.PATH_IF % 'phys-[eth1/10]'),
             base.NEXUS_IP_ADDRESS_8,
             (snipp.BODY_TRUNKVLAN % ('l1PhysIf', '', '-267')),
             base.POST],
            [(snipp.PATH_VLAN % '267'),
             base.NEXUS_IP_ADDRESS_8,
             '',
             base.DELETE]
        ],


    }


class TestCiscoNexusRestVxlanDevice(
    test_cisco_nexus_events_vxlan.TestCiscoNexusVxlanDevice):

    """Unit tests for Cisco ML2 VXLAN Nexus device driver."""

    def setUp(self):
        """Sets up mock ncclient, and switch and credentials dictionaries."""

        cfg.CONF.set_override('switch_heartbeat_time', 0, 'ml2_cisco')

        # Call Grandfather's setUp(); otherwise parent will set driver to
        # 'ncclient' instead of 'restapi'.
        super(test_cisco_nexus_events_vxlan.TestCiscoNexusVxlanDevice,
              self).setUp()
        self.mock_ncclient.reset_mock()
        self.addCleanup(self._clear_nve_db)
        self.results = TestCiscoNexusRestVxlanResults()

    def test_enable_vxlan_feature_failure(self):
        pass

    def test_disable_vxlan_feature_failure(self):
        pass

    def test_create_nve_member_failure(self):
        pass

    def test_delete_nve_member_failure(self):
        pass

    def test_nexus_vxlan_one_network_two_hosts(self):
        (super(TestCiscoNexusRestVxlanDevice, self).
            test_nexus_vxlan_one_network_two_hosts())

    def test_nexus_missing_vxlan_fields(self):
        pass

    def test_nexus_vxlan_bind_port(self):
        pass

    def test_nexus_vxlan_bind_port_no_physnet(self):
        pass

    def test_nexus_vxlan_bind_port_no_dynamic_segment(self):
        pass

    def test_nexus_vxlan_one_network(self):
        (super(TestCiscoNexusRestVxlanDevice, self).
            test_nexus_vxlan_one_network())

    def test_nexus_vxlan_two_network(self):
        (super(TestCiscoNexusRestVxlanDevice, self).
            test_nexus_vxlan_two_network())



# Copyright 2016 Cisco Systems, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron.agent.linux import dhcp

from networking_cisco import backwards_compatibility as bc


fake_tenant_id = 'aaaaaaaa-aaaa-aaaa-aaaaaaaaaaaa'
fake_subnet1_allocation_pools = dhcp.DictModel(dict(id='', start='172.9.9.2',
                                               end='172.9.9.254'))
fake_host_route = dhcp.DictModel(dict(id='', destination='40.0.1.0/24',
                                      nexthop='40.0.0.2'))
fake_subnet1 = dhcp.DictModel(dict(id='bbbbbbbb-bbbb-bbbb-bbbbbbbbbbbb',
                              network_id='12345678-1234-5678-1234567890ab',
                              cidr='172.9.9.0/24', enable_dhcp=True, name='',
                              tenant_id=fake_tenant_id,
                              gateway_ip='172.9.9.1',
                              host_routes=[fake_host_route],
                              dns_nameservers=['8.8.8.8'], ip_version=4,
                              ipv6_ra_mode=None, ipv6_address_mode=None,
                              allocation_pools=fake_subnet1_allocation_pools))

fake_subnet2_allocation_pools = dhcp.DictModel(dict(id='', start='172.9.8.2',
                                               end='172.9.8.254'))
fake_subnet2 = dhcp.DictModel(dict(id='dddddddd-dddd-dddd-dddddddddddd',
                              network_id='12345678-1234-5678-1234567890ab',
                              cidr='172.9.8.0/24', enable_dhcp=False, name='',
                              tenant_id=fake_tenant_id, gateway_ip='172.9.8.1',
                              host_routes=[], dns_nameservers=[], ip_version=4,
                              allocation_pools=fake_subnet2_allocation_pools))

fake_subnet3 = dhcp.DictModel(dict(id='bbbbbbbb-1111-2222-bbbbbbbbbbbb',
                              network_id='12345678-1234-5678-1234567890ab',
                              cidr='192.168.1.1/24', enable_dhcp=True))

fake_ipv6_subnet = dhcp.DictModel(dict(id='bbbbbbbb-1111-2222-bbbbbbbbbbbb',
                                  network_id='12345678-1234-5678-1234567890ab',
                                  cidr='2001:0db8::0/64', enable_dhcp=True,
                                  tenant_id=fake_tenant_id,
                                  gateway_ip='2001:0db8::1', ip_version=6,
                                  ipv6_ra_mode='slaac',
                                  ipv6_address_mode=None))

fake_meta_subnet = dhcp.DictModel(dict(id='bbbbbbbb-1111-2222-bbbbbbbbbbbb',
                                  network_id='12345678-1234-5678-1234567890ab',
                                  cidr='169.254.169.252/30',
                                  gateway_ip='169.254.169.253',
                                  enable_dhcp=True))

fake_fixed_ip1 = dhcp.DictModel(dict(id='', subnet_id=fake_subnet1.id,
                                ip_address='172.9.9.9'))
fake_fixed_ip2 = dhcp.DictModel(dict(id='', subnet_id=fake_subnet1.id,
                                ip_address='172.9.9.10'))
fake_fixed_ip3 = dhcp.DictModel(dict(id='', subnet_id=fake_subnet1.id,
                                ip_address='172.9.9.11'))
fake_fixed_ip4 = dhcp.DictModel(dict(id='', subnet_id=fake_subnet1.id,
                                ip_address='172.9.9.12'))
fake_fixed_ip5 = dhcp.DictModel(dict(id='', subnet_id=fake_subnet1.id,
                                ip_address='172.9.9.13'))
fake_fixed_ip6 = dhcp.DictModel(dict(id='', subnet_id=fake_subnet1.id,
                                ip_address='172.9.9.18'))
fake_fixed_ipv6 = dhcp.DictModel(dict(id='', subnet_id=fake_ipv6_subnet.id,
                                 ip_address='2001:db8::a8bb:ccff:fedd:ee99'))
fake_meta_fixed_ip = dhcp.DictModel(dict(id='', subnet=fake_meta_subnet,
                                    ip_address='169.254.169.254'))
fake_allocation_pool_subnet1 = dhcp.DictModel(dict(id='', start='172.9.9.2',
                                              end='172.9.9.254'))
fake_extra_dhcp_opts = dhcp.DictModel(dict(id='',
                                           opt_name='vendor-class-identifier',
                                           opt_value='cis'))

fake_port1 = dhcp.DictModel(dict(id='12345678-1234-aaaa-1234567890ab',
                            device_id='dhcp-12345678-1234-aaaa-1234567890ab',
                            device_owner='',
                            allocation_pools=fake_subnet1_allocation_pools,
                            mac_address='aa:bb:cc:dd:ee:ff',
                            network_id='12345678-1234-5678-1234567890ab',
                            fixed_ips=[fake_fixed_ip1, fake_fixed_ipv6],
                            extra_dhcp_opts=[fake_extra_dhcp_opts]))

fake_port2 = dhcp.DictModel(dict(id='12345678-1234-aaaa-123456789000',
                            device_id='dhcp-12345678-1234-aaaa-123456789000',
                            device_owner='',
                            mac_address='aa:bb:cc:dd:ee:99',
                            network_id='12345678-1234-5678-1234567890ab',
                            fixed_ips=[fake_fixed_ip2]))

fake_port3 = dhcp.DictModel(dict(id='12345678-1234-aaaa-123456777000',
                            device_id='dhcp-12345678-1234-aaaa-123456777000',
                            device_owner='',
                            mac_address='aa:bb:cc:dd:ff:ee',
                            network_id='12345678-1234-5678-90ab-123456777000',
                            fixed_ips=[fake_fixed_ip3, fake_fixed_ip4]))

fake_port4 = dhcp.DictModel(dict(id='12345678-1234-aaaa-123456777001',
                            device_id='dhcp-12345678-1234-aaaa-123456777001',
                            device_owner='',
                            mac_address='aa:bb:cc:ff:dd:ee',
                            network_id='12345678-1234-5678-90ab-123456777000',
                            fixed_ips=[fake_fixed_ip5, fake_fixed_ip6]))

fake_ipv6_port = dhcp.DictModel(dict(id='12345678-1234-aaaa-123456789000',
                                device_owner='',
                                mac_address='aa:bb:cc:dd:ee:99',
                                network_id='12345678-1234-5678-1234567890ab',
                                fixed_ips=[fake_fixed_ipv6]))

fake_meta_port = dhcp.DictModel(
    dict(id='12345678-1234-aaaa-1234567890ab',
         mac_address='aa:bb:cc:dd:ee:ff',
         network_id='12345678-1234-5678-1234567890ab',
         device_owner=bc.constants.DEVICE_OWNER_ROUTER_INTF,
         device_id='forzanapoli',
         fixed_ips=[fake_meta_fixed_ip]))


FAKE_NETWORK_UUID = '12345678-1234-5678-1234567890ab'
FAKE_NETWORK_DHCP_NS = "qdhcp-%s" % FAKE_NETWORK_UUID

fake_network = dhcp.NetModel(dict(id=FAKE_NETWORK_UUID,
                             tenant_id='aaaaaaaa-aaaa-aaaa-aaaaaaaaaaaa',
                             admin_state_up=True,
                             subnets=[fake_subnet1, fake_subnet2],
                             ports=[fake_port1]))

fake_network_ipv6 = dhcp.NetModel(dict(
                             id='12345678-1234-5678-1234567890ab',
                             tenant_id='aaaaaaaa-aaaa-aaaa-aaaaaaaaaaaa',
                             admin_state_up=True,
                             subnets=[fake_ipv6_subnet],
                             ports=[fake_ipv6_port]))

fake_network_ipv6_ipv4 = dhcp.NetModel(dict(
                             id='12345678-1234-5678-1234567890ab',
                             tenant_id='aaaaaaaa-aaaa-aaaa-aaaaaaaaaaaa',
                             admin_state_up=True,
                             subnets=[fake_ipv6_subnet, fake_subnet1],
                             ports=[fake_port1]))

fake_net1 = dhcp.NetModel(dict(
        id='12345678-1234-5678-90ab-1234567890ab',
        tenant_id='aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa',
        admin_state_up=True,
        subnets=[fake_subnet1, fake_ipv6_subnet],
        ports=[fake_port1]))

fake_net2 = dhcp.NetModel(dict(
        id='12345678-1234-5678-90ab-1234567890ab',
        tenant_id='aaaaaaaa-aaaa-aaaa-aaaaaaaaaaaa',
        admin_state_up=True,
        subnets=[fake_subnet1, fake_ipv6_subnet],
        ports=[fake_port1, fake_port2]))

fake_net3 = dhcp.NetModel(dict(
        id='12345678-1234-5678-90ab-123456777000',
        tenant_id='aaaaaaaa-aaaa-aaaa-aaaaaaaaaaaa',
        admin_state_up=True,
        subnets=[fake_subnet1],
        ports=[fake_port3, fake_port4]))

empty_network = dhcp.NetModel(dict(
        id='12345678-1234-5678-1234567890ab',
        tenant_id='aaaaaaaa-aaaa-aaaa-aaaaaaaaaaaa',
        admin_state_up=True,
        subnets=[fake_subnet1],
        ports=[]))

fake_meta_network = dhcp.NetModel(dict(id='12345678-1234-5678-1234567890ab',
               tenant_id='aaaaaaaa-aaaa-aaaa-aaaaaaaaaaaa',
               admin_state_up=True,
               subnets=[fake_meta_subnet],
               ports=[fake_meta_port]))



# Copyright 2015 Cisco Systems, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import uuidutils
from sqlalchemy.orm import exc
from sqlalchemy.sql import expression as expr

from neutron.db import models_v2

from neutron_lib import constants as l3_constants
from neutron_lib import exceptions as n_exc

from networking_cisco._i18n import _
from networking_cisco import backwards_compatibility as bc
from networking_cisco.backwards_compatibility import l3_const
from networking_cisco.backwards_compatibility import l3_exceptions
from networking_cisco.plugins.cisco.common import cisco_constants
from networking_cisco.plugins.cisco.db.l3 import ha_db
from networking_cisco.plugins.cisco.db.l3 import l3_models
from networking_cisco.plugins.cisco.db.l3.l3_router_appliance_db import (
    L3RouterApplianceDBMixin)
from networking_cisco.plugins.cisco.extensions import routerhostingdevice
from networking_cisco.plugins.cisco.extensions import routerrole
from networking_cisco.plugins.cisco.extensions import routertype
from networking_cisco.plugins.cisco.extensions import routertypeawarescheduler
from networking_cisco.plugins.cisco.l3 import drivers


LOG = logging.getLogger(__name__)


DEVICE_OWNER_GLOBAL_ROUTER_GW = cisco_constants.DEVICE_OWNER_GLOBAL_ROUTER_GW
HOSTING_DEVICE_ATTR = routerhostingdevice.HOSTING_DEVICE_ATTR
ROUTER_ROLE_GLOBAL = cisco_constants.ROUTER_ROLE_GLOBAL
ROUTER_ROLE_LOGICAL_GLOBAL = cisco_constants.ROUTER_ROLE_LOGICAL_GLOBAL
ROUTER_ROLE_HA_REDUNDANCY = cisco_constants.ROUTER_ROLE_HA_REDUNDANCY

TENANT_HSRP_GRP_RANGE = 1
TENANT_HSRP_GRP_OFFSET = 1064
EXT_HSRP_GRP_RANGE = 1
EXT_HSRP_GRP_OFFSET = 1064

N_ROUTER_PREFIX = 'nrouter-'
DEV_NAME_LEN = 14


class TopologyNotSupportedByRouterError(n_exc.Conflict):
    message = _("Requested topology cannot be supported by router.")


class ASR1kL3RouterDriver(drivers.L3RouterBaseDriver):

    def create_router_precommit(self, context, router_context):
        pass

    def create_router_postcommit(self, context, router_context):
        pass

    def update_router_precommit(self, context, router_context):
        pass

    def update_router_postcommit(self, context, router_context):
        # Whenever a gateway is added to, or removed from, a router hosted on
        # a hosting device, we must ensure that a global router is running
        # (for add operation) or not running (for remove operation) on that
        # hosting device.
        current = router_context.current
        if current[HOSTING_DEVICE_ATTR] is None:
            return
        e_context = context.elevated()
        if current['gw_port_id']:
            self._conditionally_add_global_router(e_context, current)
        else:
            self._conditionally_remove_global_router(
                e_context, router_context.original, True)

    def delete_router_precommit(self, context, router_context):
        pass

    def delete_router_postcommit(self, context, router_context):
        pass

    def schedule_router_precommit(self, context, router_context):
        pass

    def schedule_router_postcommit(self, context, router_context):
        # When the hosting device hosts a Neutron router with external
        # connectivity, a "global" router (modeled as a Neutron router) must
        # also run on the hosting device (outside of any VRF) to enable the
        # connectivity.
        current = router_context.current
        if current['gw_port_id'] and current[HOSTING_DEVICE_ATTR] is not None:
            self._conditionally_add_global_router(context.elevated(), current)

    def unschedule_router_precommit(self, context, router_context):
        pass

    def unschedule_router_postcommit(self, context, router_context):
        # When there is no longer any router with external gateway hosted on
        # a hosting device, the global router on that hosting device can also
        # be removed.
        current = router_context.current
        hd_id = current[HOSTING_DEVICE_ATTR]
        if current['gw_port_id'] and hd_id is not None:
            self._conditionally_remove_global_router(context.elevated(),
                                                     current)

    def add_router_interface_precommit(self, context, r_port_context):
        # Inside an ASR1k, VLAN sub-interfaces are used to connect to internal
        # neutron networks. Only one such sub-interface can be created for each
        # VLAN. As the VLAN sub-interface is added to the VRF representing the
        # Neutron router, we must only allow one Neutron router to attach to a
        # particular Neutron subnet/network.
        if (r_port_context.router_context.current[routerrole.ROUTER_ROLE_ATTR]
                == ROUTER_ROLE_HA_REDUNDANCY):
            # redundancy routers can be exempt as we check the user visible
            # routers and the request will be rejected there.
            return
        e_context = context.elevated()
        if r_port_context.current is None:
            sn = self._core_plugin.get_subnet(e_context,
                                              r_port_context.current_subnet_id)
            net_id = sn['network_id']
        else:
            net_id = r_port_context.current['network_id']
        router_id = r_port_context.router_context.current['id']
        filters = {'network_id': [net_id],
                   'device_owner': [bc.constants.DEVICE_OWNER_ROUTER_INTF]}
        for port in self._core_plugin.get_ports(e_context, filters=filters):
            device_id = port['device_id']
            if device_id is None:
                continue
            try:
                router = self._l3_plugin.get_router(e_context, device_id)
                if (router[routerrole.ROUTER_ROLE_ATTR] is None and
                        router['id'] != router_id):
                    # only a single router can connect to multiple subnets
                    # on the same internal network
                    raise TopologyNotSupportedByRouterError()
            except n_exc.NotFound:
                if self._l3_plugin.get_ha_group(e_context, device_id):
                    # Since this is a port for the HA VIP address, we can
                    # safely ignore it
                    continue
                else:
                    LOG.warning(
                        'Spurious router port %s prevents attachement from'
                        ' being performed. Try attaching again later, and '
                        'if the operation then fails again, remove the '
                        'spurious port', port['id'])
                    raise TopologyNotSupportedByRouterError()

    def add_router_interface_postcommit(self, context, r_port_context):
        pass

    def remove_router_interface_precommit(self, context, r_port_context):
        pass

    def remove_router_interface_postcommit(self, context, r_port_context):
        pass

    def create_floatingip_precommit(self, context, fip_context):
        pass

    def create_floatingip_postcommit(self, context, fip_context):
        pass

    def update_floatingip_precommit(self, context, fip_context):
        pass

    def update_floatingip_postcommit(self, context, fip_context):
        pass

    def delete_floatingip_precommit(self, context, fip_context):
        pass

    def delete_floatingip_postcommit(self, context, fip_context):
        pass

    def ha_interface_ip_address_needed(self, context, router, port,
                                       ha_settings_db, ha_group_uuid):
        if port['device_owner'] == bc.constants.DEVICE_OWNER_ROUTER_GW:
            return False
        else:
            return True

    def generate_ha_group_id(self, context, router, port, ha_settings_db,
                             ha_group_uuid):
        if port['device_owner'] in {bc.constants.DEVICE_OWNER_ROUTER_GW,
                                    DEVICE_OWNER_GLOBAL_ROUTER_GW}:
            ri_name = self._router_name(router['id'])[8:DEV_NAME_LEN]
            group_id = int(ri_name, 16) % TENANT_HSRP_GRP_RANGE
            group_id += TENANT_HSRP_GRP_OFFSET
            return group_id
        else:
            net_id_digits = port['network_id'][:6]
            group_id = int(net_id_digits, 16) % EXT_HSRP_GRP_RANGE
            group_id += EXT_HSRP_GRP_OFFSET
            return group_id

    def pre_backlog_processing(self, context):
        LOG.info('Performing pre-backlog processing')
        filters = {routerrole.ROUTER_ROLE_ATTR: [ROUTER_ROLE_GLOBAL]}
        global_routers = self._l3_plugin.get_routers(context, filters=filters)
        if not global_routers:
            LOG.debug("There are no global routers")
            return
        for gr in global_routers:
            filters = {
                HOSTING_DEVICE_ATTR: [gr[HOSTING_DEVICE_ATTR]],
                routerrole.ROUTER_ROLE_ATTR: [ROUTER_ROLE_HA_REDUNDANCY, None]
            }
            invert_filters = {'gw_port_id': [None]}
            num_rtrs = self._l3_plugin.get_routers_count_extended(
                context, filters=filters, invert_filters=invert_filters)
            LOG.debug("Global router %(name)s[%(id)s] with hosting_device "
                      "%(hd)s has %(num)d routers with gw_port set on that "
                      "device",
                      {'name': gr['name'], 'id': gr['id'],
                       'hd': gr[HOSTING_DEVICE_ATTR], 'num': num_rtrs, })
            if num_rtrs == 0:
                LOG.info(
                    "Global router %(name)s[id:%(id)s] is present for "
                    "hosting device %(hd)s but there are no tenant or "
                    "redundancy routers with gateway set on that hosting "
                    "device. Proceeding to delete global router.",
                    {'name': gr['name'], 'id': gr['id'],
                     'hd': gr[HOSTING_DEVICE_ATTR]})
                self._delete_global_router(context, gr['id'])
                filters = {
                    #TODO(bmelande): Filter on routertype of global router
                    #routertype.TYPE_ATTR: [routertype_id],
                    routerrole.ROUTER_ROLE_ATTR: [ROUTER_ROLE_LOGICAL_GLOBAL]}
                log_global_routers = self._l3_plugin.get_routers(
                    context, filters=filters)
                if log_global_routers:
                    log_global_router_id = log_global_routers[0]['id']
                    self._delete_global_router(context, log_global_router_id,
                                               logical=True)

    def post_backlog_processing(self, context):
        pass

    # ---------------- Create workflow functions -----------------

    def _conditionally_add_global_router(self, context, tenant_router):
        # We could filter on hosting device id but we don't so we get all
        # global routers for this router type. We can then use that count to
        # determine which ha priority a new global router should get.
        filters = {
            routertype.TYPE_ATTR: [tenant_router[routertype.TYPE_ATTR]],
            routerrole.ROUTER_ROLE_ATTR: [ROUTER_ROLE_GLOBAL]}
        global_routers = self._l3_plugin.get_routers(
            context, filters=filters)
        hd_to_gr_dict = {r[HOSTING_DEVICE_ATTR]: r for r in global_routers}
        hosting_device_id = tenant_router[HOSTING_DEVICE_ATTR]
        ext_nw_id = tenant_router[l3_const.EXTERNAL_GW_INFO]['network_id']
        global_router = hd_to_gr_dict.get(hosting_device_id)
        logical_global_router = self._get_logical_global_router(context,
                                                                tenant_router)
        self._conditionally_add_auxiliary_external_gateway_port(
            context, logical_global_router, ext_nw_id, tenant_router, True)
        if global_router is None:
            # must create global router on hosting device
            global_router = self._create_global_router(
                context, hosting_device_id, hd_to_gr_dict, tenant_router,
                logical_global_router)
        self._conditionally_add_auxiliary_external_gateway_port(
            context, global_router, ext_nw_id, tenant_router)
        self._l3_plugin.add_type_and_hosting_device_info(context,
                                                         global_router)
        for ni in self._l3_plugin.get_notifiers(context, [global_router]):
            if ni['notifier']:
                ni['notifier'].routers_updated(context, ni['routers'])

    def _conditionally_add_auxiliary_external_gateway_port(
            self, context, global_router, ext_net_id, tenant_router,
            provision_ha=False, port_type=DEVICE_OWNER_GLOBAL_ROUTER_GW):
        # tbe global router may or may not have an interface on the
        # external network that the tenant router uses
        filters = {
            'device_id': [global_router['id']],
            'device_owner': [port_type]}
        ext_net_port = {
            p['network_id']: p for p in
            self._core_plugin.get_ports(context, filters=filters)}
        if ext_net_id in ext_net_port:
            # already connected to the external network, called if
            # new subnets are added to the network
            aux_gw_port = self._update_auxiliary_external_gateway_port(
                context, global_router, ext_net_id, ext_net_port)
            if provision_ha:
                for subnet in aux_gw_port[ext_net_id]['fixed_ips']:
                    self._provision_port_ha(context, aux_gw_port[ext_net_id],
                                            subnet, global_router)
        else:
            # not connected to the external network, so let's fix that
            aux_gw_port = self._create_auxiliary_external_gateway_port(
                context, global_router, ext_net_id, tenant_router, port_type)
            if provision_ha:
                for subnet in aux_gw_port['fixed_ips']:
                    self._provision_port_ha(context, aux_gw_port, subnet,
                                            global_router)

    def _update_auxiliary_external_gateway_port(
            self, context, global_router, ext_net_id, port):
        # When a new subnet is added to an external network, the auxillary
        # gateway port in the global router must be updated with the new
        # subnet_id so an ip from that subnet is assigned to the gateway port
        ext_network = self._core_plugin.get_network(context, ext_net_id)
        fixed_ips = port[ext_net_id]['fixed_ips']
        # fetch the subnets the port is currently connected to
        subnet_id_list = [fixedip['subnet_id'] for fixedip in fixed_ips]
        # add the new subnet
        for subnet_id in ext_network['subnets']:
            if subnet_id not in subnet_id_list:
                fixed_ip = {'subnet_id': subnet_id}
                fixed_ips.append(fixed_ip)
                self._core_plugin.update_port(context, port[ext_net_id]['id'],
                                              ({'port': {'fixed_ips':
                                                         fixed_ips}}))
        return port

    def _create_auxiliary_external_gateway_port(
            self, context, global_router, ext_net_id, tenant_router,
            port_type=DEVICE_OWNER_GLOBAL_ROUTER_GW):
        # When a global router is connected to an external network then a
        # special type of gateway port is created on that network. Such a
        # port is called auxiliary gateway ports. It has an ip address on
        # each subnet of the external network. A (logical) global router
        # never has a traditional Neutron gateway port.
        filters = {
            'device_id': [tenant_router['id']],
            'device_owner': [l3_constants.DEVICE_OWNER_ROUTER_GW]}
        # fetch the gateway port of the *tenant* router so we can determine
        # the CIDR of that port's subnet
        gw_port = self._core_plugin.get_ports(context,
                                              filters=filters)[0]
        fixed_ips = self._get_fixed_ips_subnets(context, gw_port)
        global_router_id = global_router['id']
        aux_gw_port = self._core_plugin.create_port(context, {
            'port': {
                'tenant_id': '',  # intentionally not set
                'network_id': ext_net_id,
                'mac_address': bc.constants.ATTR_NOT_SPECIFIED,
                'fixed_ips': fixed_ips,
                'device_id': global_router_id,
                'device_owner': port_type,
                'admin_state_up': True,
                'name': ''}})
        router_port = bc.RouterPort(
                port_id=aux_gw_port['id'],
                router_id=global_router_id,
                port_type=port_type)
        context.session.add(router_port)
        return aux_gw_port

    def _create_global_router(
            self, context, hosting_device_id, hd_to_gr_dict, tenant_router,
            logical_global_router):
        r_spec = {'router': {
            # global routers are not tied to any tenant
            'tenant_id': '',
            'name': self._global_router_name(hosting_device_id),
            'admin_state_up': True}}
        global_router, r_hd_b_db = self._l3_plugin.do_create_router(
            context, r_spec, tenant_router[routertype.TYPE_ATTR], False,
            True, hosting_device_id, ROUTER_ROLE_GLOBAL)
        # make the global router a redundancy router for the logical
        # global router (which we treat as a hidden "user visible
        # router" (how's that for a contradiction of terms! :-) )
        with context.session.begin(subtransactions=True):
            ha_priority = (
                ha_db.DEFAULT_MASTER_PRIORITY -
                len(hd_to_gr_dict) * ha_db.PRIORITY_INCREASE_STEP)
            r_b_b = ha_db.RouterRedundancyBinding(
                redundancy_router_id=global_router['id'],
                priority=ha_priority,
                user_router_id=logical_global_router['id'])
            context.session.add(r_b_b)
        return global_router

    def _get_logical_global_router(self, context, tenant_router):
        # Since HA is also enabled on the global routers on each hosting device
        # those global routers need HA settings and VIPs. We represent that
        # using a Neutron router that is never instantiated/hosted. That
        # Neutron router is referred to as the "logical global" router.
        filters = {routertype.TYPE_ATTR: [tenant_router[routertype.TYPE_ATTR]],
                   routerrole.ROUTER_ROLE_ATTR: [ROUTER_ROLE_LOGICAL_GLOBAL]}
        logical_global_routers = self._l3_plugin.get_routers(
            context, filters=filters)
        if not logical_global_routers:
            # must create logical global router
            logical_global_router = self._create_logical_global_router(
                context, tenant_router)
        else:
            logical_global_router = logical_global_routers[0]
            self._update_ha_redundancy_level(context, logical_global_router, 1)
        return logical_global_router

    def _create_logical_global_router(self, context, tenant_router):
        r_spec = {'router': {
            # global routers are not tied to any tenant
            'tenant_id': '',
            'name': self._global_router_name('', logical=True),
            'admin_state_up': True,
            # set auto-schedule to false to keep this router un-hosted
            routertypeawarescheduler.AUTO_SCHEDULE_ATTR: False}}
        # notifications should never be sent for this logical router!
        logical_global_router, r_hd_b_db = (
            self._l3_plugin.do_create_router(
                context, r_spec, tenant_router[routertype.TYPE_ATTR],
                False, True, None, ROUTER_ROLE_LOGICAL_GLOBAL))
        with context.session.begin(subtransactions=True):
            r_ha_s_db = ha_db.RouterHASetting(
                router_id=logical_global_router['id'],
                ha_type=cfg.CONF.ha.default_ha_mechanism,
                redundancy_level=1,
                priority=ha_db.DEFAULT_MASTER_PRIORITY,
                probe_connectivity=False,
                probe_target=None,
                probe_interval=None)
            context.session.add(r_ha_s_db)
        return logical_global_router

    def _get_fixed_ips_subnets(self, context, gw_port):
        nw = self._core_plugin.get_network(context, gw_port['network_id'])
        subnets = [{'subnet_id': s} for s in nw['subnets']]
        return subnets

    def _provision_port_ha(self, context, ha_port, subnet, router,
                           ha_binding_db=None):
        ha_group_uuid = uuidutils.generate_uuid()
        router_id = router['id']
        with context.session.begin(subtransactions=True):
            ha_subnet_group = self._get_ha_group_by_ha_port_subnet_id(
                                   context, ha_port['id'], subnet['subnet_id'])
            if ha_subnet_group is not None:
                return
            if ha_binding_db is None:
                ha_binding_db = self._get_ha_binding(context, router_id)
            group_id = self.generate_ha_group_id(
                context, router,
                {'device_owner': DEVICE_OWNER_GLOBAL_ROUTER_GW}, ha_binding_db,
                ha_group_uuid)
            r_ha_g = ha_db.RouterHAGroup(
                id=ha_group_uuid,
                tenant_id='',
                ha_type=ha_binding_db.ha_type,
                group_identity=group_id,
                ha_port_id=ha_port['id'],
                extra_port_id=None,
                subnet_id=subnet['subnet_id'],
                user_router_id=router_id,
                timers_config='',
                tracking_config='',
                other_config='')
            context.session.add(r_ha_g)

    def _get_ha_binding(self, context, router_id):
        with context.session.begin(subtransactions=True):
            query = context.session.query(ha_db.RouterHASetting)
            query = query.filter(
                ha_db.RouterHASetting.router_id == router_id)
            return query.first()

    def _get_ha_group_by_ha_port_subnet_id(self, context, port_id, subnet_id):
        with context.session.begin(subtransactions=True):
            query = context.session.query(ha_db.RouterHAGroup)
            query = query.filter(ha_db.RouterHAGroup.ha_port_id == port_id,
                                 ha_db.RouterHAGroup.subnet_id == subnet_id)
            try:
                r_ha_g = query.one()
            except (exc.NoResultFound, exc.MultipleResultsFound):
                return
            return r_ha_g

    # ---------------- Remove workflow functions -----------------

    def _conditionally_remove_global_router(self, context, tenant_router,
                                            update_operation=False):
        filters = {routertype.TYPE_ATTR: [tenant_router[routertype.TYPE_ATTR]],
                   routerrole.ROUTER_ROLE_ATTR: [ROUTER_ROLE_GLOBAL],
                   HOSTING_DEVICE_ATTR: [tenant_router[HOSTING_DEVICE_ATTR]]}
        global_routers = self._l3_plugin.get_routers(context,
                                                     filters=filters)
        hd_to_gr_dict = {r[HOSTING_DEVICE_ATTR]: r for r in global_routers}
        if global_routers:
            global_router_id = global_routers[0]['id']
            if not tenant_router or not tenant_router[
                    l3_const.EXTERNAL_GW_INFO]:
                # let l3 plugin's periodic backlog processing take care of the
                # clean up of the global router
                return
            ext_net_id = tenant_router[l3_const.EXTERNAL_GW_INFO]['network_id']
            routertype_id = tenant_router[routertype.TYPE_ATTR]
            hd_id = tenant_router[HOSTING_DEVICE_ATTR]
            global_router = hd_to_gr_dict.get(hd_id)
            port_deleted = self._conditionally_remove_auxiliary_gateway_port(
                context, global_router_id, ext_net_id, routertype_id, hd_id,
                update_operation)
            if port_deleted is False:
                # since no auxiliary gateway port was deleted we can
                # abort no since auxiliary gateway port count cannot
                # have reached zero
                return
            filters = {
                'device_id': [global_router_id],
                'device_owner': [DEVICE_OWNER_GLOBAL_ROUTER_GW]}
            num_aux_gw_ports = self._core_plugin.get_ports_count(
                context, filters=filters)
            if num_aux_gw_ports == 0:
                # global router not needed any more so we delete it
                self._delete_global_router(context, global_router_id)
                do_notify = False
            else:
                do_notify = True
            # process logical global router to remove its port
            self._conditionally_remove_auxiliary_gateway_vip_port(
                context, ext_net_id, routertype_id)
            self._l3_plugin.add_type_and_hosting_device_info(context,
                                                             global_router)
            if do_notify is True:
                for ni in self._l3_plugin.get_notifiers(context,
                                                        [global_router]):
                    if ni['notifier']:
                        ni['notifier'].routers_updated(context, ni['routers'])

    def _conditionally_remove_auxiliary_gateway_port(
            self, context, router_id, ext_net_id, routertype_id,
            hosting_device_id, update_operation=False):
        num_rtrs = self._get_gateway_routers_count(
            context, ext_net_id, routertype_id, None, hosting_device_id)
        if ((num_rtrs <= 1 and update_operation is False) or
                (num_rtrs == 0 and update_operation is True)):
            # there are no tenant routers *on ext_net_id* that are serviced by
            # this global router so it's aux gw port can be deleted
            self._delete_auxiliary_gateway_ports(context, router_id,
                                                 ext_net_id)
            return True
        return False

    def _conditionally_remove_auxiliary_gateway_vip_port(
            self, context, ext_net_id, routertype_id):
        filters = {routertype.TYPE_ATTR: [routertype_id],
                   routerrole.ROUTER_ROLE_ATTR: [ROUTER_ROLE_LOGICAL_GLOBAL]}
        log_global_routers = self._l3_plugin.get_routers(context,
                                                         filters=filters)
        if not log_global_routers:
            return
        self._update_ha_redundancy_level(context, log_global_routers[0], -1)
        log_global_router_id = log_global_routers[0]['id']
        num_global_rtrs = self._get_gateway_routers_count(
            context, ext_net_id, routertype_id, ROUTER_ROLE_GLOBAL)
        if num_global_rtrs == 0:
            # there are no global routers *on ext_net_id* that are serviced by
            # this logical global router so it's aux gw VIP port can be deleted
            self._delete_auxiliary_gateway_ports(context, log_global_router_id,
                                                 ext_net_id)
        filters[routerrole.ROUTER_ROLE_ATTR] = [ROUTER_ROLE_GLOBAL]
        total_num_global_rtrs = self._l3_plugin.get_routers_count(
            context, filters=filters)
        if total_num_global_rtrs == 0:
            # there are no global routers left that are serviced by this
            # logical global router so it can be deleted
            self._delete_global_router(context, log_global_router_id, True)
        return False

    def _delete_auxiliary_gateway_ports(
            self, context, router_id, net_id=None,
            port_type=DEVICE_OWNER_GLOBAL_ROUTER_GW):
        filters = {
            'device_id': [router_id],
            'device_owner': [port_type]}
        if net_id is not None:
            filters['network_id'] = [net_id]
        for port in self._core_plugin.get_ports(context, filters=filters):
            try:
                self._core_plugin.delete_port(context, port['id'],
                                              l3_port_check=False)
            except (exc.ObjectDeletedError, n_exc.PortNotFound) as e:
                LOG.info('Unable to delete port for Global router '
                         '%(r_id)s. It has likely been concurrently '
                         'deleted. %(err)s', {'r_id': router_id,
                                              'err': e})

    def _delete_global_router(self, context, global_router_id, logical=False):
        # ensure we clean up any stale auxiliary gateway ports
        self._delete_auxiliary_gateway_ports(context, global_router_id)
        try:
            if logical is True:
                # We use parent class method as no special operations beyond
                # what the base implemenation does are needed for logical
                # global router
                super(L3RouterApplianceDBMixin, self._l3_plugin).delete_router(
                        context, global_router_id)
            else:
                self._l3_plugin.delete_router(
                    context, global_router_id, unschedule=False)
        except (exc.ObjectDeletedError, l3_exceptions.RouterNotFound) as e:
            g_r_type = 'Logical Global' if logical is True else 'Global'
            LOG.info('Unable to delete %(g_r_type)s router %(r_id)s. It '
                     'has likely been concurrently deleted. %(err)s',
                     {'g_r_type': g_r_type, 'r_id': global_router_id,
                     'err': e})
        except Exception as e:
            g_r_type = 'Logical Global' if logical is True else 'Global'
            LOG.debug('Failed to delete %(g_r_type)s router %(r_id). It may '
                      'have been deleted concurrently. Error details: '
                      '%(err)s',
                      {'g_r_type': g_r_type, 'r_id': global_router_id,
                       'err': e})

    def _get_gateway_routers_count(self, context, ext_net_id, routertype_id,
                                   router_role, hosting_device_id=None):
        # Determine number of routers (with routertype_id and router_role)
        # that act as gateway to ext_net_id and that are hosted on
        # hosting_device_id (if specified).
        query = context.session.query(bc.Router)
        if router_role in [None, ROUTER_ROLE_HA_REDUNDANCY]:
            # tenant router roles
            query = query.join(models_v2.Port,
                               models_v2.Port.id == bc.Router.gw_port_id)
            role_filter = expr.or_(
                l3_models.RouterHostingDeviceBinding.role == expr.null(),
                l3_models.RouterHostingDeviceBinding.role ==
                ROUTER_ROLE_HA_REDUNDANCY)
        else:
            # global and logical global routers
            query = query.join(models_v2.Port,
                               models_v2.Port.device_owner == bc.Router.id)
            role_filter = (
                l3_models.RouterHostingDeviceBinding.role == router_role)
        query = query.join(
            l3_models.RouterHostingDeviceBinding,
            l3_models.RouterHostingDeviceBinding.router_id == bc.Router.id)
        query = query.filter(
            role_filter,
            models_v2.Port.network_id == ext_net_id,
            l3_models.RouterHostingDeviceBinding.router_type_id ==
            routertype_id)
        if hosting_device_id is not None:
            query = query.filter(
                l3_models.RouterHostingDeviceBinding.hosting_device_id ==
                hosting_device_id)
        return query.count()

    # ---------------- General support functions -----------------

    def _update_ha_redundancy_level(self, context, logical_global_router,
                                    delta):
        with context.session.begin(subtransactions=True):
            log_g_router_db = self._l3_plugin._get_router(
                context, logical_global_router['id'])
            log_g_router_db.ha_settings.redundancy_level += delta
            context.session.add(log_g_router_db.ha_settings)

    def _router_name(self, router_id):
        return N_ROUTER_PREFIX + router_id

    def _global_router_name(self, hosting_device_id, logical=False):
        if logical is True:
            return cisco_constants.LOGICAL_ROUTER_ROLE_NAME
        else:
            return '%s-%s' % (cisco_constants.ROUTER_ROLE_NAME_PREFIX,
                              hosting_device_id[-cisco_constants.ROLE_ID_LEN:])

    @property
    def _core_plugin(self):
        return bc.get_plugin()

    @property
    def _l3_plugin(self):
        return bc.get_plugin(bc.constants.L3)



# Copyright (c) 2017 Cisco Systems, Inc.
# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
ML2 Nexus Driver - Helper Methods
"""

from networking_cisco import backwards_compatibility as bc


def format_interface_name(intf_type, port, ch_grp=0):
    """Method to format interface name given type, port.

    Given interface type, port, and channel-group, this
    method formats an interface name.  If channel-group is
    non-zero, then port-channel is configured.

    :param intf_type: Such as 'ethernet' or 'port-channel'
    :param port: unique identification -- 1/32 or 1
    :ch_grp: If non-zero, ignore other params and format
             port-channel<ch_grp>
    :returns: the full formatted interface name.
              ex: ethernet:1/32, port-channel:1
    """
    if ch_grp > 0:
        return 'port-channel:%s' % str(ch_grp)

    return '%s:%s' % (intf_type.lower(), port)


def split_interface_name(interface, ch_grp=0):
    """Method to split interface type, id from name.

    Takes an interface name or just interface suffix
    and returns interface type and number separately.

    :param interface: interface name or just suffix
    :param ch_grp: if non-zero, ignore interface
                   name and return 'port-channel' grp
    :returns: interface type like 'ethernet'
    :returns: returns suffix to interface name
    """

    interface = interface.lower()
    if ch_grp != 0:
        intf_type = 'port-channel'
        port = str(ch_grp)
    elif ':' in interface:
        intf_type, port = interface.split(':')
    elif interface.startswith('ethernet'):
        interface = interface.replace(" ", "")
        _, intf_type, port = interface.partition('ethernet')
    elif interface.startswith('port-channel'):
        interface = interface.replace(" ", "")
        _, intf_type, port = interface.partition('port-channel')
    else:
        intf_type, port = 'ethernet', interface

    return intf_type, port


def is_baremetal(port):
    """Identifies ironic baremetal transactions.

    There are two types of transactions.

    1. A host transaction which is dependent on host to interface mapping
       config stored in the ml2_conf.ini file. The VNIC type for this is
       'normal' which is the assumed condition.
    2. A baremetal transaction which comes from the ironic project where the
       interfaces are provided in the port transaction. In this case the
       VNIC_TYPE is 'baremetal'.
    """
    return port[bc.portbindings.VNIC_TYPE] == bc.portbindings.VNIC_BAREMETAL



# Copyright 2014 Cisco Systems, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from neutron.common import rpc as n_rpc
from oslo_config import cfg
from oslo_log import log as logging
from oslo_service import loopingcall
from oslo_utils import importutils

from networking_cisco import backwards_compatibility as bc
import networking_cisco.plugins
from networking_cisco.plugins.cisco.common import (cisco_constants as
                                                   c_constants)
from networking_cisco.plugins.cisco.db.device_manager import (
    hosting_device_manager_db as dev_mgr_db)
from networking_cisco.plugins.cisco.db.scheduler import (
    cfg_agentschedulers_db as agt_sched_db)
from networking_cisco.plugins.cisco.device_manager.rpc import (
    devices_cfgagent_rpc_cb as devices_rpc)
from networking_cisco.plugins.cisco.device_manager.rpc import (
    devmgr_rpc_cfgagent_api)
from networking_cisco.plugins.cisco.extensions import ciscocfgagentscheduler
from networking_cisco.plugins.cisco.extensions import ciscohostingdevicemanager

LOG = logging.getLogger(__name__)


class CiscoDeviceManagerPlugin(dev_mgr_db.HostingDeviceManagerMixin,
                               agt_sched_db.CfgAgentSchedulerDbMixin):
    """Implementation of Cisco Device Manager Service Plugin for Neutron.

    This class implements a (hosting) device manager service plugin that
    provides hosting device template and hosting device resources. As such
    it manages associated REST API processing. All DB functionality is
    implemented in class hosting_device_manager_db.HostingDeviceManagerMixin.
    """
    supported_extension_aliases = [
        ciscohostingdevicemanager.HOSTING_DEVICE_MANAGER_ALIAS,
        ciscocfgagentscheduler.CFG_AGENT_SCHEDULER_ALIAS]
    path_prefix = ciscocfgagentscheduler.PATH_PREFIX

    def __init__(self):
        self.setup_rpc()
        basepath = networking_cisco.plugins.__path__[0]
        ext_paths = [basepath + '/cisco/extensions']
        cp = cfg.CONF.api_extensions_path
        to_add = ""
        for ext_path in ext_paths:
            if cp.find(ext_path) == -1:
                to_add += ':' + ext_path
        if to_add != "":
            cfg.CONF.set_override('api_extensions_path', cp + to_add)
        self.cfg_agent_scheduler = importutils.import_object(
            cfg.CONF.general.configuration_agent_scheduler_driver)
        self._setup_cfg_agent_monitoring()

    def setup_rpc(self):
        # RPC support
        self.topic = c_constants.DEVICE_MANAGER_PLUGIN
        self.conn = n_rpc.create_connection()
        self.agent_notifiers[c_constants.AGENT_TYPE_CFG] = (
            devmgr_rpc_cfgagent_api.DeviceMgrCfgAgentNotifyAPI(self))
        self.endpoints = [devices_rpc.DeviceMgrCfgRpcCallback(self)]
        self.conn.create_consumer(self.topic, self.endpoints, fanout=False)
        self.conn.consume_in_threads()

    def _setup_cfg_agent_monitoring(self):
        LOG.debug('Activating periodic config agent monitor')
        self._heartbeat = loopingcall.FixedIntervalLoopingCall(
            self._check_config_agents)
        self._heartbeat.start(
            interval=cfg.CONF.general.cfg_agent_monitoring_interval)

    @property
    def _core_plugin(self):
        try:
            return self._plugin
        except AttributeError:
            self._plugin = bc.get_plugin()
            return self._plugin
